import torch
import json
import os
import codecs
from time import time
from eole.constants import CorpusTask, DefaultTokens, ModelType
from eole.inputters.dynamic_iterator import build_dynamic_dataset_iter
from eole.utils.distributed import ErrorHandler
from eole.utils.distributed_workers import spawned_infer
from eole.utils.logging import init_logger
from eole.utils.misc import get_device_type
from eole.transforms import get_transforms_cls, make_transforms, TransformPipe


class InferenceEngine(object):
    """Wrapper Class to run Inference.

    Args:
        opt: inference options
    """

    def __init__(self, config):
        self.config = config  # PredictConfig
        self.model_type = None
        torch.backends.cuda.enable_mem_efficient_sdp(True)
        torch.backends.cuda.enable_flash_sdp(False)
        torch.backends.cuda.enable_math_sdp(False)
        torch.backends.cuda.enable_cudnn_sdp(False)

    def predict_batch(self, batch):
        pass

    def _predict(self, infer_iter):
        pass

    def infer_file(self):
        """File inference. Source file must be the opt.src argument"""
        if self.config.world_size <= 1:
            infer_iter = build_dynamic_dataset_iter(
                self.config,
                self.transforms,
                self.vocabs,
                task=CorpusTask.INFER,
                device_id=self.device_id,
                model_type=self.model_type,  # patch for CT2
            )
            scores, estims, preds = self._predict(infer_iter)
        else:
            scores, estims, preds = self.infer_file_parallel()

        out_file = codecs.open(self.config.output, "w+", "utf-8")

        flatten_preds = [text for sublist in preds for text in sublist]
        flatten_scores = [score for sublist in scores for score in sublist]

        if estims is not None:
            flatten_estims = [estim for sublist in estims for estim in sublist]
        else:
            flatten_estims = [1.0 for sublist in scores for score in sublist]

        if self.config.with_score:
            if len(flatten_scores) > 0:
                out_all = [
                    pred + "\t" + str(score) + "\t" + str(estim)
                    for (pred, score, estim) in zip(flatten_preds, flatten_scores, flatten_estims)
                ]
            else:
                out_all = [str(estim) for estim in flatten_estims]
            out_file.write("\n".join(out_all) + "\n")
        else:
            out_file.write("\n".join(flatten_preds) + "\n")

        return scores, estims, preds

    def infer_list(self, src, settings={}):
        """List of strings inference `src`"""
        if self.config.world_size <= 1:
            infer_iter = build_dynamic_dataset_iter(
                self.config,
                self.transforms,
                self.vocabs,
                task=CorpusTask.INFER,
                src=src,
                device_id=self.device_id,
                model_type=self.model_type,
            )
            scores, estims, preds = self._predict(infer_iter, settings=settings)
        else:
            scores, estims, preds = self.infer_list_parallel(src, settings=settings)
        return scores, estims, preds

    def infer_file_parallel(self):
        """File inference in mulitprocessing with partitioned models."""
        raise NotImplementedError("The inference in mulitprocessing with partitioned models is not implemented.")

    def infer_list_parallel(self, src, settings={}):
        """The inference in mulitprocessing with partitioned models."""
        raise NotImplementedError("The inference in mulitprocessing with partitioned models is not implemented.")

    def _score(self, infer_iter):
        pass

    def score_file(self):
        """File scoring. Source file must be the opt.src argument"""
        if self.config.world_size <= 1:
            infer_iter = build_dynamic_dataset_iter(
                self.config,
                self.transforms,
                self.vocabs,
                task=CorpusTask.INFER,
                device_id=self.device_id,
                tgt=self.config.src,
            )
            score_results = self._score(infer_iter)
        else:
            score_results = self.score_file_parallel()
        return score_results

    def score_list(self, src):
        """List of strings scoring tgt`"""
        if self.config.world_size <= 1:
            infer_iter = build_dynamic_dataset_iter(
                self.config,
                self.transforms,
                self.vocabs,
                task=CorpusTask.INFER,
                src=src,
                tgt=src,
                device_id=self.device_id,
            )
            score_results = self._score(infer_iter)
        else:
            score_results = self.score_list_parallel(src)
        return score_results

    def terminate(self):
        pass


class InferenceEnginePY(InferenceEngine):
    """Inference engine subclass to run inference with `predict.py`.

    Args:
        opt: inference options
    """

    def __init__(self, config):
        import torch
        from eole.predict import build_predictor

        super().__init__(config)
        self.logger = init_logger(config.log_file)
        t0 = time()

        if config.world_size > 1:
            mp = torch.multiprocessing.get_context("spawn")
            # Create a thread to listen for errors in the child processes.
            self.error_queue = mp.SimpleQueue()
            self.error_handler = ErrorHandler(self.error_queue)
            self.queue_instruct = []
            self.queue_settings = []
            self.queue_result = []
            self.procs = []

            for device_id in range(config.world_size):
                self.queue_instruct.append(mp.Queue())
                self.queue_settings.append(mp.Queue())
                self.queue_result.append(mp.Queue())
                self.procs.append(
                    mp.Process(
                        target=spawned_infer,
                        args=(
                            config,
                            device_id,
                            self.error_queue,
                            self.queue_instruct[device_id],
                            self.queue_result[device_id],
                        ),
                        daemon=False,
                    )
                )
                self.procs[device_id].start()
                self.error_handler.add_child(self.procs[device_id].pid)
        else:
            if len(config.gpu_ranks) > 0:
                self.device_id = config.gpu_ranks[0]
            else:
                self.device_id = -1  # cpu

            self.predictor = build_predictor(config, self.device_id, logger=self.logger, report_score=True)
            self.transforms_cls = get_transforms_cls(config._all_transform)
            self.vocabs = self.predictor.vocabs
            self.transforms = make_transforms(config, self.transforms_cls, self.vocabs)
            self.transform_pipe = TransformPipe.build_from(self.transforms.values())
        self.logger.info("Build and loading model took %.2f sec." % (time() - t0))

    @torch.inference_mode()
    def _predict(self, infer_iter, settings={}):
        self.predictor.update_settings(**settings)
        scores, estims, preds = self.predictor._predict(
            infer_iter,
            infer_iter.transforms,
            self.config.attn_debug,
            self.config.align_debug,
        )
        return scores, estims, preds

    def _score(self, infer_iter):
        self.predictor.with_scores = True
        self.predictor.return_gold_log_probs = True
        return self.predictor._score(infer_iter)

    def score_list_parallel(self, src):
        assert self.config.world_size > 1, "World size must be greater than 1."
        for device_id in range(self.config.world_size):
            self.queue_instruct[device_id].put(("score_list", src))
        score_results = []
        for device_id in range(self.config.world_size):
            score_results.append(self.queue_result[device_id].get())
        return score_results

    def score_file_parallel(self):
        assert self.config.world_size > 1, "World size must be greater than 1."
        for device_id in range(self.config.world_size):
            self.queue_instruct[device_id].put(("score_file", self.config))
        score_results = []
        for device_id in range(self.config.world_size):
            score_results.append(self.queue_result[device_id].get())
        return score_results[0]

    def infer_file_parallel(self, settings={}):
        assert self.config.world_size > 1, "World size must be greater than 1."
        for device_id in range(self.config.world_size):
            self.queue_instruct[device_id].put(("infer_file", self.config))
            # not sure if we want a separate queue or additional info in queue_instruct
            self.queue_settings[device_id].put(settings)
        scores, estims, preds = [], [], []
        for device_id in range(self.config.world_size):
            scores.append(self.queue_result[device_id].get())
            estims.append(self.queue_result[device_id].get())
            preds.append(self.queue_result[device_id].get())
        return scores[0], estims[0], preds[0]

    def infer_list_parallel(self, src, settings={}):
        assert self.config.world_size > 1, "World size must be greater than 1."
        for device_id in range(self.config.world_size):
            self.queue_instruct[device_id].put(("infer_list", src))
            self.queue_settings[device_id].put(settings)
        scores, estims, preds = [], [], []
        for device_id in range(self.config.world_size):
            scores.append(self.queue_result[device_id].get())
            estims.append(self.queue_result[device_id].get())
            preds.append(self.queue_result[device_id].get())
        return scores[0], estims[0], preds[0]

    def terminate(self):
        if self.config.world_size > 1:
            for device_id in range(self.config.world_size):
                self.queue_instruct[device_id].put(("stop"))
                self.procs[device_id].terminate()


class InferenceEngineCT2(InferenceEngine):
    """Inference engine subclass to run inference with ctranslate2.

    Args:
        opt: inference options
    """

    def __init__(self, config, model_type=None):
        import ctranslate2
        import pyonmttok

        super().__init__(config)
        assert model_type is not None, "A model_type kwarg must be passed for CT2 models."
        self.logger = init_logger(config.log_file)
        assert self.config.world_size <= 1, "World size must be less than 1."
        if config.world_size == 1:
            self.device_id = config.gpu_ranks[0]
            self.device_index = config.gpu_ranks
            self.device = get_device_type()
        else:
            self.device_id = -1
            self.device_index = 0
            self.device = "cpu"
        self.transforms_cls = get_transforms_cls(self.config._all_transform)

        ct2_config = os.path.join(config.get_model_path() + "/ctranslate2", "config.json")
        ct2_json = json.load(open(ct2_config, "r"))
        vocabs = {}
        vocabs["specials"] = {}
        vocabs["specials"]["bos_token"] = ct2_json["bos_token"]
        vocabs["specials"]["eos_token"] = ct2_json["eos_token"]
        vocabs["specials"]["unk_token"] = ct2_json["unk_token"]
        if "pad_token" in ct2_json.keys():
            vocabs["specials"]["pad_token"] = ct2_json["pad_token"]
        else:
            vocabs["specials"]["pad_token"] = ct2_json["eos_token"]

        # Build generator or translator
        self.model_type = model_type
        if self.model_type == ModelType.DECODER:
            self.predictor = ctranslate2.Generator(
                config.get_model_path() + "/ctranslate2",
                device=self.device,
                device_index=self.device_index,
            )
            vocab_path = os.path.join(config.get_model_path() + "/ctranslate2", "vocabulary.json")
            vocab = json.load(open(vocab_path, "r"))
            src_vocab = pyonmttok.build_vocab_from_tokens(vocab)
            config.share_vocab = True
            vocabs["src"] = src_vocab
            vocabs["tgt"] = src_vocab
            vocabs["decoder_start_token"] = ""
        else:
            self.predictor = ctranslate2.Translator(
                config.get_model_path(),
                device=self.device,
                device_index=self.device_index,
            )
            vocabs["decoder_start_token"] = ct2_json["decoder_start_token"]
            if os.path.exists(config.get_model_path() + "/ctranslate2", "shared_vocabulary.json"):
                vocab = json.load(
                    open(
                        config.get_model_path() + "/ctranslate2",
                        "shared_vocabulary.json",
                        "r",
                    )
                )
                src_vocab = pyonmttok.build_vocab_from_tokens(vocab)
                config.share_vocab = True
                vocabs["src"] = src_vocab
                vocabs["tgt"] = src_vocab

            else:
                vocab_src = json.load(
                    open(
                        config.get_model_path() + "/ctranslate2",
                        "source_vocabulary.json",
                        "r",
                    )
                )
                src_vocab = pyonmttok.build_vocab_from_tokens(vocab_src)
                vocab_tgt = json.load(
                    open(
                        config.get_model_path() + "/ctranslate2",
                        "target_vocabulary.json",
                        "r",
                    )
                )
                tgt_vocab = pyonmttok.build_vocab_from_tokens(vocab_tgt)
                config.share_vocab = False
                vocabs["src"] = src_vocab
                vocabs["tgt"] = tgt_vocab

        self.vocabs = vocabs
        # Build transform pipe
        self.transforms = make_transforms(config, self.transforms_cls, self.vocabs)
        self.transforms_pipe = TransformPipe.build_from(self.transforms.values())

    def predict_batch(self, batch, config):
        input_tokens = []
        for i in range(batch["src"].size()[0]):
            start_ids = batch["src"][i, :].cpu().numpy().tolist()
            _input_tokens = [
                self.vocabs["src"].lookup_index(id)
                for id in start_ids
                if id != self.vocabs["src"].lookup_token(self.vocabs["specials"].get("pad_token", DefaultTokens.PAD))
            ]
            input_tokens.append(_input_tokens)

        if self.model_type == ModelType.DECODER:
            predicted_batch = self.predictor.generate_batch(
                start_tokens=input_tokens,
                batch_type=("examples" if config.batch_type == "sents" else "tokens"),
                max_batch_size=config.batch_size,
                beam_size=config.beam_size,
                num_hypotheses=config.n_best,
                max_length=config.max_length,
                return_scores=True,
                include_prompt_in_result=False,
                sampling_topk=config.top_k,
                sampling_topp=1 if config.top_p == 0 else config.top_p,
                sampling_temperature=config.temperature,
            )
            if self.transforms != {}:
                preds = [
                    [self.transforms_pipe.apply_reverse(nbest) for nbest in ex.sequences] for ex in predicted_batch
                ]
            else:
                preds = [[" ".join(nbest) for nbest in ex.sequences] for ex in predicted_batch]

        elif self.model_type == ModelType.ENCODER_DECODER:
            predicted_batch = self.predictor.translate_batch(
                input_tokens,
                batch_type=("examples" if config.batch_type == "sents" else "tokens"),
                max_batch_size=config.batch_size,
                beam_size=config.beam_size,
                num_hypotheses=config.n_best,
                max_decoding_length=config.max_length,
                return_scores=True,
                sampling_topk=config.top_k,
                sampling_topp=1 if config.top_p == 0 else config.top_p,
                sampling_temperature=config.temperature,
            )
            if self.transforms != {}:
                preds = [
                    [self.transforms_pipe.apply_reverse(nbest) for nbest in ex.hypothesis] for ex in predicted_batch
                ]
            else:
                preds = [[" ".join(nbest) for nbest in ex.sequences] for ex in predicted_batch]

        scores = [[nbest for nbest in ex.scores] for ex in predicted_batch]
        return scores, None, preds

    def _predict(self, infer_iter, settings={}):
        # TODO: convert settings to CT2 naming
        predictions = {}
        predictions["scores"] = []
        predictions["preds"] = []
        predictions["cid_line_number"] = []
        for batch, bucket_idx in infer_iter:
            _scores, _, _preds = self.predict_batch(batch, self.config)
            predictions["scores"] += _scores
            predictions["preds"] += _preds
            predictions["cid_line_number"] += batch["cid_line_number"]
        sorted_data = sorted(
            zip(
                predictions["cid_line_number"],
                predictions["preds"],
                predictions["scores"],
            )
        )
        sorted_predictions = {
            "cid_line_number": [item[0] for item in sorted_data],
            "preds": [item[1] for item in sorted_data],
            "scores": [item[2] for item in sorted_data],
        }
        return sorted_predictions["scores"], None, sorted_predictions["preds"]

    def _score(self, infer_iter):
        raise NotImplementedError("The scoring with InferenceEngineCT2 is not implemented.")



"""Ensemble decoding.

Decodes using multiple models simultaneously,
combining their prediction distributions by averaging.
All models in the ensemble must share a target vocabulary.
"""

import torch
import torch.nn as nn
import copy
from eole.encoders.encoder import EncoderBase
from eole.decoders.decoder import DecoderBase
from eole.models.model import EncoderDecoderModel, BaseModel


class EnsembleDecoderOutput(object):
    """Wrapper around multiple decoder final hidden states."""

    def __init__(self, model_dec_outs):
        self.model_dec_outs = tuple(model_dec_outs)

    def squeeze(self, dim=None):
        """Delegate squeeze to avoid modifying
        :func:`eole.predict.translator.Translator.predict_batch()`
        """
        return EnsembleDecoderOutput([x.squeeze(dim) for x in self.model_dec_outs])

    def __getitem__(self, index):
        return self.model_dec_outs[index]


class EnsembleSrcEmb(nn.Module):
    """Dummy SrcEmb that delegates to individual real SrcEmb."""

    def __init__(self, model_src_embs):
        super(EnsembleSrcEmb, self).__init__()
        self.model_src_embs = nn.ModuleList(model_src_embs)
        self.word_padding_idx = model_src_embs[0].word_padding_idx

    def forward(self, src):
        src_emb = [model_src_emb(src) for model_src_emb in self.model_src_embs]
        return src_emb


class EnsembleEncoder(EncoderBase):
    """Dummy Encoder that delegates to individual real Encoders."""

    def __init__(self, model_encoders):
        super(EnsembleEncoder, self).__init__()
        self.model_encoders = nn.ModuleList(model_encoders)
        if hasattr(model_encoders[0], "rope"):
            self.rope = model_encoders[0].rope

    def forward(self, emb, pad_mask=None, **kwargs):
        enc_out, enc_final_hs = zip(
            *[model_encoder(emb[i], pad_mask=pad_mask, **kwargs) for i, model_encoder in enumerate(self.model_encoders)]
        )
        return enc_out, enc_final_hs


class EnsembleTgtEmb(nn.Module):
    """Dummy TgtEmb that delegates to individual real TgtEmb."""

    def __init__(self, model_tgt_embs):
        super(EnsembleTgtEmb, self).__init__()
        self.model_tgt_embs = nn.ModuleList(model_tgt_embs)
        self.word_padding_idx = model_tgt_embs[0].word_padding_idx

    def forward(self, tgt, step=None):
        tgt_emb = [model_tgt_emb(tgt, step) for model_tgt_emb in self.model_tgt_embs]
        return tgt_emb


class EnsembleDecoder(DecoderBase):
    """Dummy Decoder that delegates to individual real Decoders."""

    def __init__(self, model_decoders):
        model_decoders = nn.ModuleList(model_decoders)
        attentional = any([dec.attentional for dec in model_decoders])
        super(EnsembleDecoder, self).__init__(attentional)
        self.model_decoders = model_decoders
        if hasattr(model_decoders[0], "rope"):
            self.rope = model_decoders[0].rope

    def forward(self, emb, enc_out=None, src_len=None, step=None, **kwargs):
        """See :func:`eole.decoders.decoder.DecoderBase.forward()`."""
        # src_len is a single tensor shared between all models.
        # This assumption will not hold if Translator is modified
        # to calculate src_len as something other than the length
        # of the input.
        dec_outs, attns = zip(
            *[
                model_decoder(
                    emb[i],
                    enc_out=None if enc_out is None else enc_out[i],
                    src_len=src_len,
                    step=step,
                    **kwargs,
                )
                for i, model_decoder in enumerate(self.model_decoders)
            ]
        )
        if attns[0]["std"] is not None:
            mean_attns = self.combine_attns(attns)
        else:
            mean_attns = attns
        return EnsembleDecoderOutput(dec_outs), mean_attns

    def combine_attns(self, attns):
        result = {}
        for key in attns[0].keys():
            result[key] = torch.stack([attn[key] for attn in attns if attn[key] is not None]).mean(0)
        return result

    def init_state(self, enc_out=None, src=None, enc_final_hs=None):
        """See :obj:`RNNDecoderBase.init_state()`"""
        for i, model_decoder in enumerate(self.model_decoders):
            if enc_out is not None:
                enc_out_i = enc_out[i]
            else:
                enc_out_i = None
            if enc_final_hs is not None:
                enc_final_hs_i = enc_final_hs[i]
            else:
                enc_final_hs_i = None
            model_decoder.init_state(src=src, enc_out=enc_out_i, enc_final_hs=enc_final_hs_i)

    def map_state(self, fn):
        for model_decoder in self.model_decoders:
            model_decoder.map_state(fn)


class EnsembleGenerator(nn.Module):
    """
    Dummy Generator that delegates to individual real Generators,
    and then averages the resulting target distributions.
    """

    def __init__(self, model_generators, raw_probs=False):
        super(EnsembleGenerator, self).__init__()
        self.model_generators = nn.ModuleList(model_generators)
        self._raw_probs = raw_probs

    def forward(self, hidden, attn=None, src_map=None):
        """
        Compute a distribution over the target dictionary
        by averaging distributions from models in the ensemble.
        All models in the ensemble must share a target vocabulary.
        """
        distributions = torch.stack(
            [mg(h) if attn is None else mg(h, attn, src_map) for h, mg in zip(hidden, self.model_generators)]
        )
        if self._raw_probs:
            return torch.log(torch.exp(distributions).mean(0))
        else:
            return distributions.mean(0)


class EnsembleModel(EncoderDecoderModel):
    """Dummy EncoderDecoderModel wrapping individual real EncoderDecoderModels."""

    def __init__(self, models, raw_probs=False):
        src_emb = EnsembleSrcEmb([model.src_emb for model in models])
        encoder = EnsembleEncoder([model.encoder for model in models])
        tgt_emb = EnsembleTgtEmb([model.tgt_emb for model in models])
        decoder = EnsembleDecoder([model.decoder for model in models])
        hidden_size = models[0].hidden_size
        super(EnsembleModel, self).__init__(
            encoder=encoder,
            decoder=decoder,
            src_emb=src_emb,
            tgt_emb=tgt_emb,
            hidden_size=hidden_size,
        )
        self.generator = EnsembleGenerator([model.generator for model in models], raw_probs)
        self.models = nn.ModuleList(models)


def load_test_model(config, device_id=0):
    """Read in multiple models for ensemble."""
    shared_vocabs = None
    shared_model_config = None
    models = []
    config2 = copy.deepcopy(config)
    for i, model_path in enumerate(config.model_path):
        config2.model_path = [config.model_path[i]]
        vocabs, model, model_config = BaseModel.load_test_model(config2, device_id, model_path=model_path)
        if shared_vocabs is None:
            shared_vocabs = vocabs
        else:
            assert (
                shared_vocabs["src"].tokens_to_ids == vocabs["src"].tokens_to_ids
            ), "Ensemble models must use the same vocabs "
        models.append(model)
        if shared_model_config is None:
            shared_model_config = model_config
    ensemble_model = EnsembleModel(models, config.avg_raw_probs)
    return shared_vocabs, ensemble_model, shared_model_config



from eole.transforms import register_transform
from .transform import Transform, TransformConfig
from pydantic import Field
from typing import List
from eole.utils.logging import logger
import os
import urllib.request
import regex as re


class CleanConfig(TransformConfig):
    src_eq_tgt: bool | None = Field(default=False, description="Remove ex src==tgt")
    same_char: bool | None = Field(default=False, description="Remove ex with same char more than 4 times")
    same_word: bool | None = Field(default=False, description="Remove ex with same word more than 3 times")
    scripts_ok: List[str] | None = Field(
        default=["Latin", "Common"],
        description="list of unicodata scripts accepted",
    )
    scripts_nok: List[str] | None = Field(
        default=[],
        description="list of unicodata scripts not accepted",
    )
    src_tgt_ratio: float | None = Field(default=2.0, description="ratio between src and tgt")
    avg_tok_min: float | None = Field(default=3.0, description="average length of tokens min")
    avg_tok_max: float | None = Field(default=20.0, description="average length of tokens max")
    langid: List[str] | None = Field(default=[], description="list of languages accepted")


@register_transform(name="clean")
class CleanTransform(Transform):
    """
    Clean examples according to rules
    """

    config_model = CleanConfig

    def __init__(self, config):
        super().__init__(config)

    def _parse_config(self):
        self.src_eq_tgt = self.config.src_eq_tgt
        self.same_char = self.config.same_char
        self.same_word = self.config.same_word
        self.scripts_ok = self.config.scripts_ok
        self.scripts_nok = self.config.scripts_nok
        self.src_tgt_ratio = self.config.src_tgt_ratio
        self.avg_tok_min = self.config.avg_tok_min
        self.avg_tok_max = self.config.avg_tok_max
        self.langid = self.config.langid
        assert self.scripts_ok == [] or self.scripts_nok == [], "Choose either scripts to be included or excluded"

    # this whole logic could probably be improved
    @staticmethod
    def _get_param(corpus, param, def_val):
        """Get param string of a `corpus`."""
        if "clean" in getattr(corpus, "transforms", []):
            value = getattr(corpus, param, def_val)
            clean = value
        else:
            clean = None
        return clean

    @classmethod
    def get_config_dict(cls, config, param, def_val):
        """Get clean settings correspond to corpus in `opts`."""
        clean_dict = {}
        # normalize dict src/tgt for each dataset
        if hasattr(config, "data"):
            for c_name, corpus in config.data.items():
                clean = cls._get_param(corpus, param, def_val)
                if clean is not None:
                    logger.debug(f"Get {config} for {c_name}: {clean}")
                    clean_dict[c_name] = clean

        return clean_dict

    def warm_up(self, vocabs=None):
        super().warm_up(None)
        import fasttext

        self.src_eq_tgt_dict = self.get_config_dict(self.full_config, "src_eq_tgt", True)
        self.same_char_dict = self.get_config_dict(self.full_config, "same_char", True)
        self.same_word_dict = self.get_config_dict(self.full_config, "same_word", True)
        self.scripts_ok_dict = self.get_config_dict(self.full_config, "scripts_ok", ["Latin", "Common"])
        self.scripts_nok_dict = self.get_config_dict(self.full_config, "scripts_nok", [])
        self.src_tgt_ratio_dict = self.get_config_dict(self.full_config, "src_tgt_ratio", 2)
        self.avg_tok_min_dict = self.get_config_dict(self.full_config, "avg_tok_min", 3)
        self.avg_tok_max_dict = self.get_config_dict(self.full_config, "avg_tok_max", 20)
        self.langid_dict = self.get_config_dict(self.full_config, "langid", [])
        fasttext_loc = f"{os.path.dirname(os.path.abspath(__file__))}/lid.176.ftz"

        if not os.path.exists(fasttext_loc):
            urllib.request.urlretrieve(
                "https://dl.fbaipublicfiles.com/" + "fasttext/supervised-models/lid.176.ftz",
                fasttext_loc,
            )
        self.id_func = fasttext.load_model(fasttext_loc)

    def batch_apply(self, batch, is_train=False, stats=None, **kwargs):
        """Convert source and target examples to doc level segments."""

        def _id(string):
            res = self.id_func.predict(string, k=1)
            res = res[0][0].replace("__label__", "")
            return res

        trf_batch = []

        for ex, _, cid in batch:
            if self.scripts_ok_dict[cid]:
                ok_regex = "[^" + "".join(r"\p{%s}" % sc for sc in self.scripts_ok_dict[cid]) + "]"

            if self.scripts_nok_dict[cid]:
                nok_regex = "[" + "".join(r"\p{%s}" % sc for sc in self.scripts_nok_dict[cid]) + "]"

            src_str = " ".join(ex["src"])
            if len(src_str) == 0:
                # print("src empty")
                continue
            if self.same_char_dict[cid] and re.search(r"([^0-9])\1{3}", src_str):
                # print("too many same char in src")
                continue
            if self.same_word_dict[cid] and re.search(r"(\ .*|.*\ )\1{2}", src_str):
                # print("too many same word in src")
                continue
            if len(src_str) / len(ex["src"]) < self.avg_tok_min_dict[cid]:
                # print("avg token min", len(src_str) / len(ex['src']))
                continue
            if len(src_str) / len(ex["src"]) > self.avg_tok_max_dict[cid]:
                # print("avg token max", len(src_str) / len(ex['src']))
                continue

            if self.scripts_ok_dict[cid] and re.search(ok_regex, src_str):
                # print("text does not fully belong to wanted script")
                continue
            if self.scripts_nok_dict[cid] and re.search(nok_regex, src_str):
                # print("Some text belong to unwanted scripts")
                continue

            if self.langid_dict[cid] != [] and _id(src_str) not in self.langid_dict[cid]:
                # print("langid does not match", _id(src_str))
                continue

            if ex["tgt"] is not None:
                tgt_str = " ".join(ex["tgt"])
                if self.src_eq_tgt_dict[cid] and src_str == tgt_str:
                    # print("src = tgt")
                    continue
                if len(tgt_str) == 0:
                    # print("tgt empty")
                    continue
                if (len(ex["src"]) + 1) / (len(ex["tgt"]) + 1) > self.src_tgt_ratio_dict[cid] or (
                    len(ex["src"]) + 1
                ) / (len(ex["tgt"]) + 1) < (1 / self.src_tgt_ratio_dict[cid]):
                    # print("src / tgt ratio ", len(src_str) / len(tgt_str))
                    continue
                if self.same_char_dict[cid] and re.search(r"([^0-9])\1{3}", tgt_str):
                    # print("too many same char in tgt")
                    continue
                if self.same_word_dict[cid] and re.search(r"(\ .*|.*\ )\1{2}", tgt_str):
                    # print("too many same word in tgt")
                    continue
                if len(tgt_str) / len(ex["tgt"]) < self.avg_tok_min_dict[cid]:
                    # print("avg token min", len(tgt_str) / len(ex['tgt']))
                    continue
                if len(tgt_str) / len(ex["tgt"]) > self.avg_tok_max_dict[cid]:
                    # print("avg token max", len(tgt_str) / len(ex['tgt']))
                    continue

                if self.scripts_ok_dict[cid] and re.search(ok_regex, tgt_str):
                    # print("text does not fully belong to wanted script")
                    continue
                if self.scripts_nok_dict[cid] and re.search(nok_regex, tgt_str):
                    # print("Some text belong to unwanted scripts")
                    continue

                if self.langid_dict[cid] != [] and _id(tgt_str) not in self.langid_dict[cid]:
                    # print("langid does not match", _id(tgt_str))
                    continue

            trf_batch.append((ex, self, cid))

        return trf_batch



import torch
from typing import List, Literal
from pydantic import Field, model_validator, field_validator, computed_field

from eole.config.common import RunningConfig, LoRaConfig, QuantizeConfig
from eole.config.config import Config, get_config_dict


class DecodingConfig(Config):
    beam_size: int = Field(default=5, description="Beam size.")
    ratio: float = Field(default=-0.0, description="Ratio based beam stop condition.")  # is the minus sign useful here?
    top_k: int = Field(
        default=0,
        description="Set this to -1 to do random sampling from full distribution. "
        "Set this to value k>1 to do random sampling restricted to "
        "the k most likely next tokens. Set this to 1 to use argmax.",
    )
    top_p: float = Field(
        default=0.0,
        description="Probability for top-p/nucleus sampling. "
        "Restrict tokens to the most likely until the cumulated probability "
        "is over p. In range [0,1]. (https://arxiv.org/abs/1904.09751)",
        ge=0.0,
        lte=1.0,
    )
    temperature: float = Field(
        default=1.0,
        description="If doing random sampling, divide the logits by this " "before computing softmax during decoding.",
    )
    length_penalty: Literal["avg", "wu", "none"] = Field(default="avg", description="Length penalty to use.")
    alpha: float = Field(default=1.0, description="Length penalty parameter (higher = longer generation)")
    coverage_penalty: Literal["none", "wu", "summary"] = Field(
        default="none",
        description="Coverage penalty to use. Only available in beam search.",
    )
    beta: float = Field(default=-0.0, description="Coverage penalty parameter.")
    stepwise_penalty: bool = Field(
        default=False,
        description="Apply coverage penalty at every decoding step. Helpful for summary penalty.",
    )
    min_length: int = Field(default=0, description="Minimum prediction length.", ge=0)
    max_length: int = Field(default=250, description="Maximum prediction length.")
    max_length_ratio: float = Field(
        default=2,
        description="Maximum prediction length ratio. For European languages, "
        "2 is large enough, for target Asian charageters, "
        "need to increase to 2-3, for special languages (Burmese, Amharic) to 10.",
        ge=1,
    )  # we might want to validate this against min_length
    block_ngram_repeat: int = Field(default=0, description="Block repetition of ngrams during decoding.")
    ignore_when_blocking: List[str] = Field(
        default=[],
        description="Ignore these strings when blocking repeats. " "You want to block sentence delimiters.",
    )
    replace_unk: bool = Field(
        default=False,
        description="Replace the generated UNK tokens with the source token "
        "that had the highest attention weight. If phrase_table is provided, "
        "it will lok up the identified source token and give the "
        "corresponding target token. If it is not provided "
        "(or the identified source token does not exist in the table), "
        "then it will copy the source token.",
    )  # deprecated?
    ban_unk_token: bool = Field(
        default=False,
        description="Prevent unk token generation by setting unk probability to 0.",
    )
    phrase_table: str = Field(
        default="",
        description="If phrase_table is provided (with replace_unk), it will look up "
        "the identified source token and give the corresponding target token.",
    )  # deprecated?
    n_best: int = Field(default=1, description="Output the n_best decoded sentences.")
    # added while testing, might be put elsewhere
    dump_beam: str = Field(
        default="", description="File to dump beam information to."
    )  # not sure it's still working/useful
    verbose: bool = Field(default=False, description="Print scores and predictions for each input.")
    with_score: bool = Field(default=False, description="Add a tab separated score to each output.")
    estim_only: bool = Field(default=False, description="Process the input to estimator only (no decoder).")
    attn_debug: bool = Field(default=False, description="Print best attn for each word.")
    align_debug: bool = Field(default=False, description="Print best align for each word.")


# in legacy opts, decoding config is separated (probably to be used elsewhere)
class InferenceConfig(RunningConfig, DecodingConfig, LoRaConfig, QuantizeConfig):

    model_config = get_config_dict()
    model_config["arbitrary_types_allowed"] = True  # to allow torch.dtype

    report_align: bool = Field(default=False, description="Report alignment for each translation.")
    gold_align: bool = Field(
        default=False,
        description="Report alignment between source and gold target. "
        "Useful to test the performance of learnt alignments.",
    )
    report_time: bool = Field(default=False, description="Report some translation time metrics.")
    fuse_kvq: bool = Field(default=False, description="Fuse K, V, Q Linear layers into a single KVQ in Self Attn.")
    fuse_gate: bool = Field(
        default=False, description="Fuse gate_up_proj and up_proj Linear layers into a single Linear."
    )
    profile: bool = Field(default=False, description="Report pytorch profiling stats.")
    batch_size: int = Field(default=30, description="Batch size.")
    batch_type: Literal["sents", "tokens"] = Field(default="sents", description="Batch grouping for batch size.")
    avg_raw_probs: bool = Field(
        default=False,
        description="If set, during ensembling scores from different models will be combined "
        "by averaging their raw probabilities and then taking the log. "
        "Otherwise, the log probabilities will be averaged directly. "
        "Necessary for models whose output layers can assign zero probability.",
    )
    data_type: str | None = "text"  # deprecated? hopefully will change with input streams logic
    chat_template: str | None = None
    optional_eos: List[str] | None = Field(
        default=[],
        description="Optional EOS tokens that would stop generation, e.g. <|eot_id|> for Llama3",
    )

    def get_model_path(self):
        return self.model_path[0]

    @field_validator("model_path")
    @classmethod
    def _validate_model_path(cls, v: str | List[str]) -> List[str]:
        if isinstance(v, str):
            v = [v]
        return v

    @model_validator(mode="after")
    def _validate_running_config(self):
        # super()._validate_running_config()
        if self.gold_align:
            assert self.report_align, "-report_align should be enabled with -gold_align"
            assert not self.replace_unk, "-replace_unk option can not be used with -gold_align enabled"
            assert self.tgt, "-tgt should be specified with -gold_align"
        # originally in validate_translate_opts_dynamic, not sure why
        # self.__dict__["share_vocab"] = False
        if self.compute_dtype == torch.int8:
            assert self.gpu < 0, "Dynamic 8-bit quantization is not supported on GPU"
        return self

    @computed_field
    @property
    def storage_dtype(self) -> torch.dtype:
        """
        Deduce which dtype to use for main model parameters.
        """
        if self.compute_dtype == torch.int8:
            return torch.float32
        return self.compute_dtype



# -*- coding: utf-8 -*-
import os
from math import ceil
import codecs
import pyonmttok
from eole.constants import DefaultTokens


def build_vocab(config, specials):
    """Build vocabs dict to be stored in the checkpoint
        based on vocab files having each line [token, count]
    Args:
        opt: src_vocab, tgt_vocab
    Return:
        vocabs: {'src': pyonmttok.Vocab, 'tgt': pyonmttok.Vocab,
                 'decoder_start_token': DefaultTokens.BOS
                }
    """
    vocabs = {}
    vocabs["specials"] = {
        "unk_token": config.unk_token,
        "pad_token": config.pad_token,
        "bos_token": config.bos_token,
        "eos_token": config.eos_token,
    }

    def _pad_vocab_to_multiple(vocab, multiple):
        vocab_size = len(vocab)
        if vocab_size % multiple == 0:
            return vocab
        target_size = int(ceil(vocab_size / multiple)) * multiple
        for i in range(target_size - vocab_size):
            vocab.add_token(DefaultTokens.VOCAB_PAD + str(i))
        return vocab

    default_specials = list(vocabs["specials"].values())

    src_vocab = _read_vocab_file(config.src_vocab, config.src_words_min_frequency)

    src_specials = [item for item in (default_specials + specials["src"]) if item not in src_vocab]

    if DefaultTokens.SEP in src_specials and ("<0x0A>" in src_vocab or "Ċ" in src_vocab):
        # this is hack: if the special separator ｟newline｠is returned because of the
        # "docify" transform.get_specials we don't add it if the corresponding newline code
        # is already included in the sentencepiece or BPE-with-gpt2-pretok.
        src_specials.remove(DefaultTokens.SEP)

    src_vocab = pyonmttok.build_vocab_from_tokens(
        src_vocab, maximum_size=config.src_vocab_size, special_tokens=src_specials
    )
    src_vocab.default_id = src_vocab[config.unk_token]
    if src_vocab.default_id >= len(src_vocab):
        src_vocab.default_id = 0  # patch that assigns OOV to id=0 when UNK does not exist

    if config.vocab_size_multiple > 1:
        src_vocab = _pad_vocab_to_multiple(src_vocab, config.vocab_size_multiple)
    vocabs["src"] = src_vocab
    if config.share_vocab:
        vocabs["tgt"] = src_vocab
    else:
        tgt_vocab = _read_vocab_file(config.tgt_vocab, config.tgt_words_min_frequency)
        tgt_specials = [item for item in (default_specials + specials["tgt"]) if item not in tgt_vocab]
        if DefaultTokens.SEP in tgt_specials and ("<0x0A>" in tgt_vocab or "Ċ" in src_vocab):
            tgt_specials.remove(DefaultTokens.SEP)
        tgt_vocab = pyonmttok.build_vocab_from_tokens(
            tgt_vocab, maximum_size=config.tgt_vocab_size, special_tokens=tgt_specials
        )
        tgt_vocab.default_id = tgt_vocab[DefaultTokens.UNK]
        if tgt_vocab.default_id >= len(tgt_vocab):
            tgt_vocab.default_id = 0  # patch that assigns OOV to id=0 when UNK does not exist
        if config.vocab_size_multiple > 1:
            tgt_vocab = _pad_vocab_to_multiple(tgt_vocab, config.vocab_size_multiple)
        vocabs["tgt"] = tgt_vocab

    vocabs["decoder_start_token"] = config.decoder_start_token
    return vocabs


def _read_vocab_file(vocab_path, min_count):
    """Loads a vocabulary from the given path.

    Args:
        vocab_path (str): Path to utf-8 text file containing vocabulary.
            Each token should be on a line, may followed with a count number
            seperate by space if `with_count`. No extra whitespace is allowed.
        min_count (int): retains only tokens with min_count frequency.
    """

    if not os.path.exists(vocab_path):
        raise RuntimeError("Vocabulary not found at {}".format(vocab_path))
    else:
        with codecs.open(vocab_path, "rb") as f:
            lines = [line.decode("utf-8") for line in f.read().split(b"\n")]
            lines = lines[:-1]

            first_line = lines[0].split(None, 1)
            has_count = len(first_line) == 2 and first_line[-1].strip().isdigit()
            if has_count:
                vocab = []
                for line in lines:
                    if int(line.split(None, 1)[1]) >= min_count:
                        vocab.append(line.split(None, 1)[0])
            else:
                vocab = lines
            return vocab


def vocabs_to_dict(vocabs):
    """
    Convert a dict of pyonmttok vocabs
    into a plain text dict to be saved in the checkpoint
    """
    vocabs_dict = {}
    vocabs_dict["src"] = vocabs["src"].ids_to_tokens
    vocabs_dict["tgt"] = vocabs["tgt"].ids_to_tokens

    if "decoder_start_token" in vocabs.keys():
        vocabs_dict["decoder_start_token"] = vocabs["decoder_start_token"]
    else:
        vocabs_dict["decoder_start_token"] = DefaultTokens.BOS
    if "specials" in vocabs.keys():
        vocabs_dict["specials"] = vocabs["specials"]
    return vocabs_dict


def dict_to_vocabs(vocabs_dict):
    """
    Convert a dict formatted vocabs (as stored in a checkpoint)
    into a dict of pyonmttok vocabs objects.
    """
    vocabs = {}
    if "decoder_start_token" in vocabs_dict.keys():
        vocabs["decoder_start_token"] = vocabs_dict["decoder_start_token"]
    else:
        vocabs["decoder_start_token"] = DefaultTokens.BOS
    vocabs["src"] = pyonmttok.build_vocab_from_tokens(vocabs_dict["src"])
    if vocabs["src"].default_id >= len(vocabs["src"]):
        vocabs["src"].default_id = 0  # patch that assigns OOV to id=0
    if vocabs_dict["src"] == vocabs_dict["tgt"]:
        vocabs["tgt"] = vocabs["src"]
    else:
        vocabs["tgt"] = pyonmttok.build_vocab_from_tokens(vocabs_dict["tgt"])
        if vocabs["tgt"].default_id >= len(vocabs["src"]):
            vocabs["tgt"].default_id = 0  # patch that assigns OOV to id=0
    if "specials" in vocabs_dict.keys():
        vocabs["specials"] = vocabs_dict["specials"]

    return vocabs



"""
Attention Entropy Computation Utilities

This module provides functions to compute entropy of attention weights
during training and inference, which can be useful for analyzing
attention patterns and model behavior.
"""

import torch
from typing import Dict, Optional, Union, List


def compute_attention_entropy(
    attention_weights: torch.Tensor, mask: Optional[torch.Tensor] = None, eps: float = 1e-8
) -> torch.Tensor:
    """
    Compute entropy of attention weights.

    Entropy is computed as: H(A) = -Σ(A_ij * log(A_ij))
    where A_ij are the attention weights.

    Args:
        attention_weights (torch.Tensor): Attention weights tensor of shape
            [batch_size, num_heads, seq_len_q, seq_len_k] or
            [batch_size, seq_len_q, seq_len_k]
        mask (torch.Tensor, optional): Mask tensor to ignore padded positions.
            Should have the same shape as attention_weights or be broadcastable.
        eps (float): Small epsilon value to avoid log(0). Default: 1e-8

    Returns:
        torch.Tensor: Entropy values. Shape depends on input:
            - If input has heads dimension: [batch_size, num_heads]
            - Otherwise: [batch_size]
    """
    # Ensure attention weights are positive and normalized
    attn = torch.clamp(attention_weights, min=eps)

    # Apply mask if provided
    if mask is not None:
        # Convert mask to same dtype as attention weights
        mask = mask.to(attn.dtype)
        # Zero out masked positions
        attn = attn * mask
        # Renormalize after masking
        if len(attn.shape) == 4:  # [batch, heads, seq_q, seq_k]
            attn = attn / (attn.sum(dim=-1, keepdim=True) + eps)
        else:  # [batch, seq_q, seq_k]
            attn = attn / (attn.sum(dim=-1, keepdim=True) + eps)

    # Compute entropy: H = -Σ(p * log(p))
    log_attn = torch.log(attn + eps)
    entropy = -torch.sum(attn * log_attn, dim=-1)  # Sum over key dimension

    # Average over query dimension to get per-head or per-batch entropy
    entropy = torch.mean(entropy, dim=-1)

    return entropy


def compute_attention_entropy_from_dict(
    attns_dict: Dict[str, Union[torch.Tensor, List[torch.Tensor]]],
    attention_types: Optional[List[str]] = None,
    layer_indices: Optional[List[int]] = None,
    eps: float = 1e-8,
) -> Dict[str, torch.Tensor]:
    """
    Compute attention entropy from attention dictionary returned by model.

    Args:
        attns_dict (Dict): Dictionary of attention weights from model forward pass.
            Keys are attention types (e.g., 'std', 'self', 'context').
            Values are either single tensors or lists of tensors (one per layer).
        attention_types (List[str], optional): Which attention types to compute entropy for.
            If None, computes for all available types.
        layer_indices (List[int], optional): Which layer indices to include.
            If None, includes all layers.
        eps (float): Small epsilon value to avoid log(0).

    Returns:
        Dict[str, torch.Tensor]: Dictionary with entropy values.
            Keys are in format "{attention_type}_layer_{idx}" or just "{attention_type}"
            if single tensor. Values are entropy tensors.
    """
    entropy_dict = {}

    if attention_types is None:
        attention_types = list(attns_dict.keys())

    for attn_type in attention_types:
        if attn_type not in attns_dict:
            continue

        attn_data = attns_dict[attn_type]

        if isinstance(attn_data, list):
            # Multiple layers
            for layer_idx, layer_attn in enumerate(attn_data):
                if layer_indices is None or layer_idx in layer_indices:
                    if layer_attn is not None:
                        entropy = compute_attention_entropy(layer_attn, eps=eps)
                        key = f"{attn_type}_layer_{layer_idx}"
                        entropy_dict[key] = entropy
        else:
            # Single tensor
            if attn_data is not None:
                entropy = compute_attention_entropy(attn_data, eps=eps)
                entropy_dict[attn_type] = entropy

    return entropy_dict


def aggregate_attention_entropy(
    entropy_dict: Dict[str, torch.Tensor], aggregation_method: str = "mean"
) -> torch.Tensor:
    """
    Aggregate attention entropy values across different attention types/layers.

    Args:
        entropy_dict (Dict[str, torch.Tensor]): Dictionary of entropy values.
        aggregation_method (str): How to aggregate. Options: "mean", "max", "min".

    Returns:
        torch.Tensor: Aggregated entropy value(s).
    """
    if not entropy_dict:
        return torch.tensor(0.0)

    # Stack all entropy values
    entropy_values = []
    for entropy_tensor in entropy_dict.values():
        if entropy_tensor.numel() > 0:
            # If tensor has multiple dimensions (e.g., per head), take mean
            entropy_values.append(entropy_tensor.mean())

    if not entropy_values:
        return torch.tensor(0.0)

    entropy_stack = torch.stack(entropy_values)

    if aggregation_method == "mean":
        return entropy_stack.mean()
    elif aggregation_method == "max":
        return entropy_stack.max()
    elif aggregation_method == "min":
        return entropy_stack.min()
    else:
        raise ValueError(f"Unknown aggregation method: {aggregation_method}")


def compute_batch_attention_entropy(
    attns_dict: Dict[str, Union[torch.Tensor, List[torch.Tensor]]],
    attention_types: Optional[List[str]] = None,
    layer_indices: Optional[List[int]] = None,
    aggregation_method: str = "mean",
    eps: float = 1e-8,
) -> float:
    """
    Compute and aggregate attention entropy for a batch.

    This is a convenience function that combines entropy computation and aggregation.

    Args:
        attns_dict (Dict): Dictionary of attention weights from model forward pass.
        attention_types (List[str], optional): Which attention types to compute entropy for.
        layer_indices (List[int], optional): Which layer indices to include.
        aggregation_method (str): How to aggregate entropy values.
        eps (float): Small epsilon value to avoid log(0).

    Returns:
        float: Aggregated attention entropy value for the batch.
    """
    entropy_dict = compute_attention_entropy_from_dict(attns_dict, attention_types, layer_indices, eps)

    aggregated_entropy = aggregate_attention_entropy(entropy_dict, aggregation_method)

    return aggregated_entropy.item()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from eole.inference_engine import InferenceEnginePY, InferenceEngineCT2
from eole.constants import ModelType
from argparse import ArgumentParser
from eole.utils.misc import use_gpu, set_random_seed
from torch.profiler import profile, record_function, ProfilerActivity
from eole.config.run import PredictConfig
from eole.bin import register_bin
from eole.bin.run import RunBin


def model_type(config) -> ModelType:
    if config.decoder is None:
        return ModelType.ENCODER
    elif config.encoder is None:
        return ModelType.DECODER
    else:
        return ModelType.ENCODER_DECODER


def predict(config):
    set_random_seed(config.seed, use_gpu(config))
    if config.engine == "eole":
        engine = InferenceEnginePY(config)
    elif config.engine == "ct2":
        engine = InferenceEngineCT2(config, model_type(config.model))
    else:
        raise ValueError("You need to use --engine with 'eole' or 'ct2'")
    _, _, _ = engine.infer_file()
    engine.terminate()


@register_bin(name="predict")
class Predict(RunBin):
    config_class = PredictConfig
    require_config = False

    @classmethod
    def run(cls, args):
        config = cls.build_config(args)

        if config.profile:
            with profile(
                activities=[ProfilerActivity.CPU, ProfilerActivity.CUDA],
                profile_memory=True,
                with_stack=True,
            ) as prof:
                with record_function("Predict"):
                    predict(config)
            print(prof.key_averages().table(sort_by="cuda_time_total", row_limit=40))
        else:
            predict(config)


def _get_parser():
    # Patch to allow autodoc with sphinx-argparse
    parser = ArgumentParser()
    Predict.add_args(parser)
    return parser



import torch
from eole.predict import penalties
from eole.predict.decode_strategy import DecodeStrategy
import warnings


class BeamSearchBase(DecodeStrategy):
    """Generation beam search.

    Note that the attributes list is not exhaustive. Rather, it highlights
    tensors to document their shape. (Since the state variables' "batch"
    size decreases as beams finish, we denote this axis with a B rather than
    ``batch_size``).

    Args:
        beam_size (int): Number of beams to use (see base ``parallel_paths``).
        batch_size (int): See base.
        pad (int): See base.
        bos (int): See base.
        eos (int): See base.
        unk (int): See base.
        start (int): See base.
        n_best (int): Don't stop until at least this many beams have
            reached EOS.
        global_scorer (eole.predict.GNMTGlobalScorer): Scorer instance.
        min_length (int): See base.
        max_length (int): See base.
        return_attention (bool): See base.
        block_ngram_repeat (int): See base.
        exclusion_tokens (set[int]): See base.

    Attributes:
        _batch_offset (LongTensor): Shape ``(B,)``.
        _beam_offset (LongTensor): Shape ``(batch_size x beam_size,)``.
        alive_seq (LongTensor): See base.
        topk_log_probs (FloatTensor): Shape ``(B, beam_size,)``. These
            are the scores used for the topk operation.
        src_len (LongTensor): Lengths of encodings. Used for
            masking attentions.
        select_indices (LongTensor or NoneType): Shape
            ``(B x beam_size,)``. This is just a flat view of the
            ``_batch_index``.
        topk_scores (FloatTensor): Shape
            ``(B, beam_size)``. These are the
            scores a sequence will receive if it finishes.
        topk_ids (LongTensor): Shape ``(B, beam_size)``. These are the
            word indices of the topk predictions.
        _batch_index (LongTensor): Shape ``(B, beam_size)``.
        _prev_penalty (FloatTensor or NoneType): Shape
            ``(B, beam_size)``. Initialized to ``None``.
        _coverage (FloatTensor or NoneType): Shape
            ``(1, B x beam_size, inp_seq_len)``.
        hypotheses (list[list[Tuple[Tensor]]]): Contains a tuple
            of score (float), sequence (long), and attention (float or None).
    """

    def __init__(
        self,
        beam_size,
        batch_size,
        pad,
        bos,
        eos,
        unk,
        start,
        n_best,
        global_scorer,
        min_length,
        max_length,
        return_attention,
        block_ngram_repeat,
        exclusion_tokens,
        stepwise_penalty,
        ratio,
        ban_unk_token,
        add_estimator=False,
    ):
        super(BeamSearchBase, self).__init__(
            pad,
            bos,
            eos,
            unk,
            start,
            batch_size,
            beam_size,
            global_scorer,
            min_length,
            block_ngram_repeat,
            exclusion_tokens,
            return_attention,
            max_length,
            ban_unk_token,
            add_estimator,
        )
        # beam parameters
        self.beam_size = beam_size
        self.n_best = n_best
        self.ratio = ratio
        if add_estimator:
            self.num_hyp = self.beam_size
        else:
            self.num_hyp = self.n_best

        # beam state
        self._batch_offset = torch.arange(batch_size, dtype=torch.long)

        self.select_indices = None
        self.done = False
        # "global state" of the old beam
        self._prev_penalty = None
        self._coverage = None

        self._stepwise_cov_pen = stepwise_penalty and self.global_scorer.has_cov_pen
        self._vanilla_cov_pen = not stepwise_penalty and self.global_scorer.has_cov_pen
        self._cov_pen = self.global_scorer.has_cov_pen

        self.src_len = None

    def initialize(self, *args, **kwargs):
        raise NotImplementedError

    def initialize_(self, enc_out, device, target_prefix):
        super(BeamSearchBase, self).initialize(device, target_prefix)
        self.best_scores = [-1e10 for _ in range(self.batch_size)]
        self._beam_offset = torch.arange(
            0,
            self.batch_size * self.beam_size,
            step=self.beam_size,
            dtype=torch.long,
            device=device,
        )
        self.topk_log_probs = (
            torch.tensor([0.0] + [float("-inf")] * (self.beam_size - 1), device=device)
            .repeat(self.batch_size)
            .reshape(self.batch_size, self.beam_size)
        )
        # buffers for the topk scores and 'backpointer'
        self.topk_scores = torch.empty((self.batch_size, self.beam_size), dtype=torch.float, device=device)
        # MPS doesn't support torch.isin() in Torch 2.3
        # Avoiding need to CPU fallback by adding alternative implementation
        # Can be removed when Torch 2.4 is supported
        self._is_finished_list = (
            self._is_finished_list_mps if (device is not None and device.type == "mps") else self._is_finished_list_isin
        )
        """
        self.topk_ids = torch.empty(
            (self.batch_size, self.beam_size), dtype=torch.long, device=device
        )

        self._batch_index = torch.empty(
            [self.batch_size, self.beam_size], dtype=torch.long, device=device
        )
        """

    @property
    def current_predictions(self):
        return self.alive_seq[:, -1]

    @property
    def current_backptr(self):
        # for testing
        return self.select_indices.view(self.batch_size, self.beam_size).fmod(self.beam_size)

    @property
    def batch_offset(self):
        return self._batch_offset

    def _pick(self, log_probs):
        """Take a token pick decision for a step.

        Args:
            log_probs (FloatTensor): (B * beam_size, vocab_size)

        Returns:
            topk_scores (FloatTensor): (B, beam_size)
            topk_ids (LongTensor): (B, beam_size)
        """
        vocab_size = log_probs.size(-1)
        # maybe fix some prediction at this step by modifying log_probs
        log_probs = self.target_prefixing(log_probs)

        # Flatten probs into a list of possibilities.
        curr_scores = log_probs.reshape(-1, self.beam_size * vocab_size)

        topk_scores, topk_ids = torch.topk(curr_scores, self.beam_size, dim=-1)
        # after this we get topk_ids between 0 and beam_size*vocab_size
        # topk_ids // vocab_size => indice in beam
        # topk_ids % vocab_size => true vocab indice

        return topk_scores, topk_ids

    def beams_non_finished(self, i, topk_scores_list, predictions, attention, step):
        # using lists instead of tensors for topk_scores and is_finished make things faster
        if any(self.is_finished_list[i]):
            b = self._batch_offset[i]
            # Store finished hypotheses for this example in the batch.
            for j in [
                k for k, fin in enumerate(self.is_finished_list[i]) if fin
            ]:  # Beam level: finished beam j in example i of batch
                if self.ratio > 0:
                    s = topk_scores_list[i][j] / (step + 1)
                    self.best_scores[b] = max(s, self.best_scores[b])
                self.hypotheses[b].append(
                    (
                        topk_scores_list[i][j],
                        predictions[i, j, 1:],  # Ignore start_token.
                        attention[i, j, :, : self.src_len[i]] if attention is not None else None,
                    )
                )

            # End condition is the top beam finished and we can return
            # n_best hypotheses.
            if self.ratio > 0:
                pred_len = self.src_len[i] * self.ratio
                finish_flag = ((topk_scores_list[i][0] / pred_len) <= self.best_scores[b]) or all(
                    self.is_finished_list[i]
                )
            else:
                # early stop when top beam is finished
                finish_flag = self.is_finished_list[i][0]

            if finish_flag and len(self.hypotheses[b]) >= self.num_hyp:
                self.hypotheses[b] = sorted(self.hypotheses[b], key=lambda x: x[0], reverse=True)
                for score, pred, attn in self.hypotheses[b][: self.num_hyp]:
                    self.scores[b].append(score)
                    self.predictions[b].append(pred)  # ``(batch, n_best,)``
                    self.attention[b].append(attn if attn is not None else [])
                return False
            else:
                return True
        else:
            return True

    def update_finished(self):
        # Penalize beams that finished.
        _B_old = self.topk_log_probs.shape[0]
        step = self.alive_seq.shape[-1]  # 1 greater than the step in advance
        # this is required to pursue finished beams in non finished batches
        self.topk_log_probs.masked_fill_(
            torch.tensor(self.is_finished_list, device=self.topk_log_probs.device),
            -65504,
        )
        predictions = self.alive_seq.view(_B_old, self.beam_size, step)
        attention = (
            self.alive_attn.view(_B_old, self.beam_size, step - 1, self.alive_attn.size(-1))
            if self.alive_attn is not None
            else None
        )

        topk_scores_list = self.topk_scores.tolist()
        non_finished_batch = [
            i
            for i in range(len(self.is_finished_list))
            if self.beams_non_finished(i, topk_scores_list, predictions, attention, step)
        ]

        non_finished = torch.tensor(non_finished_batch)
        # If all sentences are predicted, no need to go further.
        if len(non_finished) == 0:
            self.done = True
            return

        _B_new = non_finished.shape[0]
        self.remove_finished_batches(_B_new, _B_old, non_finished, predictions, attention, step)

        # reset the selection for the next step
        self.select_indices = self._batch_index.view(_B_new * self.beam_size)
        self.src_len = self.src_len[self.select_indices]
        self.maybe_update_target_prefix(self.select_indices)

    def remove_finished_batches(self, _B_new, _B_old, non_finished, predictions, attention, step):
        # Remove finished batches for the next step.
        self._batch_offset = self._batch_offset[non_finished]  # CPU
        non_finished = non_finished.to(self.topk_log_probs.device)
        self.topk_log_probs = self.topk_log_probs[non_finished]
        self._batch_index = self._batch_index[non_finished]
        self.alive_seq = predictions[non_finished].view(-1, self.alive_seq.size(-1))

        if self.alive_attn is not None:
            inp_seq_len = self.alive_attn.size(-1)
            self.alive_attn = attention[non_finished].view(_B_new * self.beam_size, step - 1, inp_seq_len)
            if self._cov_pen:
                self._coverage = self._coverage.view(_B_old, self.beam_size, 1, inp_seq_len)[non_finished].view(
                    _B_new * self.beam_size, 1, inp_seq_len
                )
                if self._stepwise_cov_pen:
                    self._prev_penalty = self._prev_penalty[non_finished]

    def advance(self, log_probs, attn):
        vocab_size = log_probs.size(-1)

        # using integer division to get an integer _B without casting
        _B = log_probs.shape[0] // self.beam_size

        if self._stepwise_cov_pen and self._prev_penalty is not None:
            self.topk_log_probs += self._prev_penalty
            self.topk_log_probs -= self.global_scorer.cov_penalty(self._coverage + attn, self.global_scorer.beta).view(
                _B, self.beam_size
            )

        # force the output to be longer than self.min_length
        step = len(self)
        self.ensure_min_length(log_probs)
        self.ensure_unk_removed(log_probs)
        # Multiply probs by the beam probability.
        log_probs += self.topk_log_probs.view(_B * self.beam_size, 1)

        # if the sequence ends now, then the penalty is the current
        # length + 1, to include the EOS token
        length_penalty = self.global_scorer.length_penalty(step + 1, alpha=self.global_scorer.alpha)
        if length_penalty != 1:
            curr_scores = log_probs / length_penalty
        else:
            curr_scores = log_probs

        # Avoid any direction that would repeat unwanted ngrams
        self.block_ngram_repeats(curr_scores)

        # Pick up candidate token by curr_scores
        self.topk_scores, self.topk_ids = self._pick(curr_scores)

        # Recover log probs.
        # Length penalty is just a scalar. It doesn't matter if it's applied
        # before or after the topk.
        self.topk_log_probs = self.topk_scores * length_penalty

        # Resolve beam origin and map to batch index flat representation.
        self._batch_index = self.topk_ids // vocab_size + self._beam_offset[:_B].unsqueeze(1)
        self.select_indices = self._batch_index.view(_B * self.beam_size)
        self.topk_ids %= vocab_size

        # Append last prediction to reordered alive sequence
        self.alive_seq = torch.cat(
            [
                self.alive_seq[self.select_indices],
                self.topk_ids.view(_B * self.beam_size, 1),
            ],
            -1,
        )

        self.maybe_update_forbidden_tokens()

        if self.return_attention or self._cov_pen:
            current_attn = attn[self.select_indices]
            if step == 1:
                self.alive_attn = current_attn
                # update global state (step == 1)
                if self._cov_pen:  # coverage penalty
                    self._prev_penalty = torch.zeros_like(self.topk_log_probs)
                    self._coverage = current_attn
            else:
                self.alive_attn = self.alive_attn[self.select_indices]
                self.alive_attn = torch.cat([self.alive_attn, current_attn], 1)
                # update global state (step > 1)
                if self._cov_pen:
                    self._coverage = self._coverage[self.select_indices]
                    self._coverage += current_attn
                    self._prev_penalty = self.global_scorer.cov_penalty(
                        self._coverage, beta=self.global_scorer.beta
                    ).view(_B, self.beam_size)

        if self._vanilla_cov_pen:
            # shape: (batch_size x beam_size, 1)
            cov_penalty = self.global_scorer.cov_penalty(self._coverage, beta=self.global_scorer.beta)
            self.topk_scores -= cov_penalty.view(_B, self.beam_size).float()

        self.is_finished_list = self._is_finished_list()
        self.ensure_max_length()

    def _is_finished_list_isin(self):
        return torch.isin(self.topk_ids, self.eos_t).tolist()

    def _is_finished_list_mps(self):
        return (self.topk_ids.unsqueeze(1) == self.eos_t).sum(dim=1).bool().tolist()


class BeamSearch(BeamSearchBase):
    """
    Beam search for seq2seq/encoder-decoder models
    """

    def initialize(self, enc_out, src_len, device=None, target_prefix=None):
        """Initialize for decoding.
        Repeat src objects `beam_size` times.
        """

        (fn_tile, enc_out, target_prefix) = self.initialize_tile(enc_out, src_len, target_prefix)
        if device is None:
            device = self.get_device_from_enc_out(enc_out)

        self.eos_t = torch.tensor(self.eos).to(device)
        super(BeamSearch, self).initialize_(enc_out, device, target_prefix)

        return fn_tile, enc_out


class BeamSearchLM(BeamSearchBase):
    """
    Beam search for language/decoder only models
    """

    def initialize(self, src, src_len, device=None, target_prefix=None):
        """Initialize for decoding.
        Repeat src objects `beam_size` times.
        """
        (fn_tile, _, target_prefix) = self.initialize_tile(None, src_len, target_prefix)
        if device is None:
            device = src.device

        self.eos_t = torch.tensor(self.eos).to(device)
        super(BeamSearchLM, self).initialize_(
            None,
            device=device,
            target_prefix=target_prefix,
        )

        return fn_tile, src

    def advance(self, log_probs, attn):
        super(BeamSearchLM, self).advance(log_probs, attn)

        # in LM task src_len is associated with currently generated src
        # and therefore needs to follow the generation
        self.src_len += 1


class GNMTGlobalScorer(object):
    """NMT re-ranking.

    Args:
       alpha (float): Length parameter.
       beta (float):  Coverage parameter.
       length_penalty (str): Length penalty strategy.
       coverage_penalty (str): Coverage penalty strategy.

    Attributes:
        alpha (float): See above.
        beta (float): See above.
        length_penalty (callable): See :class:`penalties.PenaltyBuilder`.
        coverage_penalty (callable): See :class:`penalties.PenaltyBuilder`.
        has_cov_pen (bool): See :class:`penalties.PenaltyBuilder`.
        has_len_pen (bool): See :class:`penalties.PenaltyBuilder`.
    """

    @classmethod
    def from_config(cls, config):
        return cls(config.alpha, config.beta, config.length_penalty, config.coverage_penalty)

    def __init__(self, alpha, beta, length_penalty, coverage_penalty):
        self._validate(alpha, beta, length_penalty, coverage_penalty)
        self.alpha = alpha
        self.beta = beta
        penalty_builder = penalties.PenaltyBuilder(coverage_penalty, length_penalty)
        self.has_cov_pen = penalty_builder.has_cov_pen
        # Term will be subtracted from probability
        self.cov_penalty = penalty_builder.coverage_penalty

        self.has_len_pen = penalty_builder.has_len_pen
        # Probability will be divided by this
        self.length_penalty = penalty_builder.length_penalty

    @classmethod
    def _validate(cls, alpha, beta, length_penalty, coverage_penalty):
        # these warnings indicate that either the alpha/beta
        # forces a penalty to be a no-op, or a penalty is a no-op but
        # the alpha/beta would suggest otherwise.
        if length_penalty is not None and alpha == 0.0:
            warnings.warn("Using length penalty with alpha==0 " "is equivalent to using length penalty none.")
        if coverage_penalty is None or coverage_penalty == "none":
            if beta != 0:
                warnings.warn("Non-default `beta` with no coverage penalty. " "`beta` has no effect.")
        else:
            # using some coverage penalty
            if beta == 0.0:
                warnings.warn(
                    "Non-default coverage penalty with beta==0 " "is equivalent to using coverage penalty none."
                )



import torch
from torch.nn.utils.rnn import pad_sequence
from eole.predict.inference import Inference
from eole.constants import ModelType
from eole.predict.greedy_search import GreedySearchLM
from eole.predict.beam_search import BeamSearchLM
from eole.utils.misc import tile
from time import time


class GeneratorLM(Inference):
    @classmethod
    def validate_task(cls, task):
        if task != ModelType.DECODER:
            raise ValueError(f"GeneratorLM does not support task {task}." f" Tasks supported: {ModelType.DECODER}")

    def _align_forward(self, batch, predictions):
        """
        For a batch of input and its prediction, return a list of batch predict
        alignment src indice Tensor in size ``(batch, n_best,)``.
        """
        raise NotImplementedError

    def predict_batch(self, batch, attn_debug, scoring=False):
        """Predict a batch of sentences."""
        max_length = 0 if scoring else self.max_length
        with torch.no_grad():
            if self.top_k != 0 or self.top_p != 0:
                decode_strategy = GreedySearchLM(
                    pad=self._tgt_pad_idx,
                    bos=self._tgt_bos_idx,
                    eos=self._tgt_eos_idx,
                    unk=self._tgt_unk_idx,
                    start=self._tgt_start_with,
                    n_best=self.n_best,
                    batch_size=len(batch["srclen"]),
                    global_scorer=self.global_scorer,
                    min_length=self.min_length,
                    max_length=max_length,
                    block_ngram_repeat=self.block_ngram_repeat,
                    exclusion_tokens=self._exclusion_idxs,
                    return_attention=attn_debug or self.replace_unk,
                    temperature=self.temperature,
                    top_k=self.top_k,
                    top_p=self.top_p,
                    beam_size=self.beam_size,
                    ban_unk_token=self.ban_unk_token,
                    add_estimator=self.add_estimator,
                )
            else:
                # TODO: support these blacklisted features
                assert not self.dump_beam
                decode_strategy = BeamSearchLM(
                    self.beam_size,
                    batch_size=len(batch["srclen"]),
                    pad=self._tgt_pad_idx,
                    bos=self._tgt_bos_idx,
                    eos=self._tgt_eos_idx,
                    unk=self._tgt_unk_idx,
                    start=self._tgt_start_with,
                    n_best=self.n_best,
                    global_scorer=self.global_scorer,
                    min_length=self.min_length,
                    max_length=max_length,
                    return_attention=attn_debug or self.replace_unk,
                    block_ngram_repeat=self.block_ngram_repeat,
                    exclusion_tokens=self._exclusion_idxs,
                    stepwise_penalty=self.stepwise_penalty,
                    ratio=self.ratio,
                    ban_unk_token=self.ban_unk_token,
                    add_estimator=self.add_estimator,
                )
            return self._predict_batch_with_strategy(batch, decode_strategy)

    @classmethod
    def split_src_to_prevent_padding(cls, src, src_len):
        min_len_batch = torch.min(src_len).item()
        target_prefix = None
        if min_len_batch > 0 and min_len_batch < src.size(1):
            target_prefix = src[:, min_len_batch:]
            src = src[:, :min_len_batch]
            src_len[:] = min_len_batch
        return src, src_len, target_prefix

    def tile_to_beam_size_after_initial_step(self, fn_tile, log_probs):
        if fn_tile is not None:
            log_probs = fn_tile(log_probs)
            self.model.decoder.map_state(fn_tile)
            log_probs = log_probs[:, -1, :]
        return log_probs

    def _predict_batch_with_strategy(self, batch, decode_strategy):
        """Predict a batch of sentences step by step using cache.

        Args:
            batch: a batch of sentences, yield by data iterator.
            decode_strategy (DecodeStrategy): A decode strategy to use for
                generate prediction step by step.

        Returns:
            results (dict): The prediction results.
        """

        # (0) Prep the components of the search.
        parallel_paths = decode_strategy.parallel_paths  # beam_size
        batch_size = len(batch["srclen"])

        # (1) split src into src and target_prefix to avoid padding.
        src = batch["src"]
        src_len = batch["srclen"]

        if batch["left_pad"]:
            target_prefix = None
        else:
            src, src_len, target_prefix = self.split_src_to_prevent_padding(src, src_len)

        # (2) init decoder
        self.model.decoder.init_state(src=src)
        gold_score, gold_log_probs = self._gold_score(batch, None, src_len, None, batch_size, src)

        # (3) prep decode_strategy. Possibly repeat src objects.
        (fn_tile, src) = decode_strategy.initialize(
            src,
            src_len,
            target_prefix=target_prefix,
        )
        if not self.estim_only:
            # (4) Begin decoding step by step:
            if self.report_time:
                torch.cuda.synchronize()
                beg_time = time()

            for step in range(decode_strategy.max_length):
                decoder_input = src if step == 0 else decode_strategy.current_predictions.view(-1, 1)

                log_probs, attn = self._decode_and_generate(
                    decoder_input,
                    None,
                    src_len=decode_strategy.src_len,
                    step=step if step == 0 else step + max(src_len.tolist()),
                    left_pad=batch["left_pad"],
                    images=batch.get("images", None),
                )

                if step == 0:
                    log_probs = self.tile_to_beam_size_after_initial_step(fn_tile, log_probs)

                decode_strategy.advance(log_probs, attn)
                any_finished = any([any(sublist) for sublist in decode_strategy.is_finished_list])
                if any_finished:
                    decode_strategy.update_finished()
                    if decode_strategy.done:
                        break
                select_indices = decode_strategy.select_indices

                if parallel_paths > 1 or any_finished:
                    # select indexes in model state/cache
                    self.model.decoder.map_state(lambda state: state[select_indices])

                if self.report_time and step == 0:
                    torch.cuda.synchronize()
                    self.step0_time.append(time() - beg_time)

        if self.add_estimator:
            # Prepare estimator input = decoder out of each pred with initial enc_out
            if not self.estim_only:
                dec_in = [item for sublist in decode_strategy.predictions for item in sublist]
                src = tile(src, parallel_paths)
                concat_seq = [torch.cat((s, d), dim=0) for s, d in zip(src, dec_in)]
                # make padding left sided
                concat_seq = [x.flip(dims=[0]) for x in concat_seq]
                concat_seq = pad_sequence(concat_seq, batch_first=True, padding_value=self._tgt_pad_idx)
                concat_seq = concat_seq.flip(dims=[1])
                # remove eos
                dec_in = concat_seq[:, :-1]
            else:
                dec_in = src
                parallel_paths = 1
            tgt_pad_mask = dec_in.eq(self._tgt_pad_idx).unsqueeze(1)  # [B, T_tgt]
            emb = self.model.tgt_emb(dec_in)
            self.model.decoder._disable_cache()
            dec_out, _ = self.model.decoder(
                emb,
                enc_out=None,
                return_attn=False,
                tgt_pad_mask=tgt_pad_mask,
            )
            if self.estimator_type == "average":
                pad_mask = ~dec_in.eq(self._tgt_pad_idx)
                in_estim = (dec_out * pad_mask.unsqueeze(-1).float()).sum(dim=1) / pad_mask.sum(
                    dim=1, keepdim=True
                ).float()
            elif self.estimator_type == "last_token":
                in_estim = dec_out[:, -1, :]
            else:
                raise ValueError("Decoder only model should use average or last token estimator")
            estim = self.model.estimator(in_estim.to(dec_out.dtype)).squeeze(-1)
            estim = [
                [estim[i].item() for i in range(j, j + parallel_paths)] for j in range(0, len(estim), parallel_paths)
            ]
        else:
            estim = [[1.0 for _ in range(self.beam_size)] for _ in range(batch_size)]

        return self.report_results(
            gold_score,
            gold_log_probs,
            batch,
            batch_size,
            decode_strategy,
            estim,
        )

    def _score_target(self, batch, enc_out, src_len):
        src = batch["src"]
        src_len = batch["srclen"]
        tgt = batch["tgt"]

        log_probs, attn = self._decode_and_generate(
            src,
            None,
            src_len=src_len,
        )

        log_probs[:, :, self._tgt_pad_idx] = 0
        tgt = tgt.unsqueeze(2)
        gold_log_probs = log_probs.gather(2, tgt).squeeze(-1)
        gold_scores = gold_log_probs.sum(dim=1).view(-1)

        if self.return_gold_log_probs:
            return gold_scores, gold_log_probs

        return gold_scores, None



import unittest

from eole.transforms.bart import word_start_finder
from eole.utils.alignment import subword_map_by_joiner, subword_map_by_spacer
from eole.constants import SubwordMarker


class TestWordStartFinder(unittest.TestCase):
    def test_word_start_naive(self):
        word_start_finder_fn = word_start_finder(ignore_subword=True)
        data_in = [
            "however",
            ",",
            "according",
            "to",
            "the",
            "logs",
            ",",
            "she",
            "is",
            "hard",
            "-",
            "working",
            ".",
        ]  # noqa: E501
        true_out = [
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
        ]  # noqa: E501
        out = word_start_finder_fn(data_in)
        self.assertEqual(out, true_out)

    def test_word_start_joiner(self):
        word_start_finder_fn = word_start_finder(is_joiner=True)
        data_in = [
            "however",
            "￭,",
            "according",
            "to",
            "the",
            "logs",
            "￭,",
            "she",
            "is",
            "hard",
            "￭-￭",
            "working",
            "￭.",
        ]  # noqa: E501
        true_out = [
            True,
            False,
            True,
            True,
            True,
            True,
            False,
            True,
            True,
            True,
            False,
            False,
            False,
        ]  # noqa: E501
        out = word_start_finder_fn(data_in)
        self.assertEqual(out, true_out)

    def test_word_start_spacer(self):
        word_start_finder_fn = word_start_finder()
        data_in = [
            "▁however",
            ",",
            "▁according",
            "▁to",
            "▁the",
            "▁logs",
            ",",
            "▁she",
            "▁is",
            "▁hard",
            "-",
            "working",
            ".",
        ]  # noqa: E501
        true_out = [
            True,
            False,
            True,
            True,
            True,
            True,
            False,
            True,
            True,
            True,
            False,
            False,
            False,
        ]  # noqa: E501
        out = word_start_finder_fn(data_in)
        self.assertEqual(out, true_out)
        # no dummy prefix
        no_dummy = [
            "however",
            ",",
            "▁according",
            "▁to",
            "▁the",
            "▁logs",
            ",",
            "▁she",
            "▁is",
            "▁hard",
            "-",
            "working",
            ".",
        ]  # noqa: E501
        no_dummy_out = word_start_finder_fn(no_dummy)
        self.assertEqual(no_dummy_out, true_out)


class TestSubwordGroup(unittest.TestCase):
    def test_subword_group_joiner(self):
        data_in = [
            "however",
            "￭,",
            "according",
            "to",
            "the",
            "logs",
            "￭,",
            "she",
            "is",
            "hard",
            "￭-￭",
            "working",
            "￭.",
        ]  # noqa: E501
        true_out = [0, 0, 1, 2, 3, 4, 4, 5, 6, 7, 7, 7, 7]
        out = subword_map_by_joiner(data_in, marker=SubwordMarker.JOINER)
        self.assertEqual(out, true_out)

    def test_subword_group_joiner_with_case_markup(self):
        data_in = [
            "｟mrk_case_modifier_C｠",
            "however",
            "￭,",
            "according",
            "to",
            "the",
            "logs",
            "￭,",
            "｟mrk_begin_case_region_U｠",
            "she",
            "is",
            "hard",
            "￭-￭",
            "working",
            "｟mrk_end_case_region_U｠",
            "￭.",
        ]  # noqa: E501
        true_out = [0, 0, 0, 1, 2, 3, 4, 4, 5, 5, 6, 7, 7, 7, 7, 7]
        out = subword_map_by_joiner(data_in, marker=SubwordMarker.JOINER)
        self.assertEqual(out, true_out)

    def test_subword_group_joiner_with_case_markup_advanced(self):
        data_in = [
            "｟mrk_case_modifier_C｠",
            "dummy",
            "text",
            "｟mrk_case_modifier_C｠",
            "1￭",
            "h￭",
            "k",
            "｟mrk_begin_case_region_U｠",
            "th￭",
            "｟mrk_end_case_region_U｠",
            "n",
            "more",
            "dummy",
            "text",
        ]  # noqa: E501
        true_out = [0, 0, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 5, 6]
        out = subword_map_by_joiner(data_in, marker=SubwordMarker.JOINER)
        self.assertEqual(out, true_out)

    def test_subword_group_joiner_prior_tokenization(self):
        data_in = [
            "｟mrk_case_modifier_C｠",
            "how￭",
            "ever",
            "￭,",
            "according",
            "to",
            "the",
            "logs",
            "￭,",
            "｟mrk_begin_case_region_U｠",
            "she",
            "is",
            "hard",
            "￭-￭",
            "working",
            "｟mrk_end_case_region_U｠",
            "￭.",
        ]  # noqa: E501
        original_data_in = [
            "However",
            "￭,",
            "according",
            "to",
            "the",
            "logs",
            "￭,",
            "SHE",
            "IS",
            "HARD-WORKING",
            "￭.",
        ]  # noqa: E501
        true_out = [0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 7, 8, 9, 9, 9, 9, 10]  # noqa: E501
        out = subword_map_by_joiner(data_in, marker=SubwordMarker.JOINER, original_subwords=original_data_in)
        self.assertEqual(out, true_out)

    def test_subword_group_joiner_prior_tokenization_harder(self):
        data_in = [
            "｟mrk_case_modifier_C｠",
            "how￭",
            "ever",
            "￭,",
            "according",
            "to",
            "the",
            "logs",
            "￭,",
            "｟mrk_begin_case_region_U｠",
            "she",
            "is",
            "hard",
            "￭-￭",
            "working",
            "｟mrk_end_case_region_U｠",
            "￭.",
        ]  # noqa: E501
        original_data_in = [
            "｟mrk_case_modifier_C｠",
            "how￭",
            "ever",
            "￭,",
            "according",
            "to",
            "the",
            "logs",
            "￭,",
            "｟mrk_begin_case_region_U｠",
            "she",
            "is",
            "hard",
            "￭-￭",
            "working",
            "｟mrk_end_case_region_U｠",
            "￭.",
        ]  # noqa: E501
        true_out = [
            0,
            1,
            2,
            3,
            4,
            5,
            6,
            7,
            8,
            9,
            10,
            11,
            12,
            13,
            14,
            15,
            16,
        ]  # noqa: E501
        out = subword_map_by_joiner(data_in, marker=SubwordMarker.JOINER, original_subwords=original_data_in)
        self.assertEqual(out, true_out)

    def test_subword_group_joiner_with_new_joiner(self):
        data_in = [
            "｟mrk_case_modifier_C｠",
            "however",
            "￭",
            ",",
            "according",
            "to",
            "the",
            "logs",
            "￭",
            ",",
            "｟mrk_begin_case_region_U｠",
            "she",
            "is",
            "hard",
            "￭",
            "-",
            "￭",
            "working",
            "｟mrk_end_case_region_U｠",
            "￭",
            ".",
        ]  # noqa: E501
        true_out = [
            0,
            0,
            0,
            0,
            1,
            2,
            3,
            4,
            4,
            4,
            5,
            5,
            6,
            7,
            7,
            7,
            7,
            7,
            7,
            7,
            7,
        ]  # noqa: E501
        out = subword_map_by_joiner(data_in, marker=SubwordMarker.JOINER)
        self.assertEqual(out, true_out)

    def test_subword_group_naive(self):
        data_in = [
            "however",
            ",",
            "according",
            "to",
            "the",
            "logs",
            ",",
            "she",
            "is",
            "hard",
            "-",
            "working",
            ".",
        ]  # noqa: E501
        true_out = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]
        out = subword_map_by_joiner(data_in, marker=SubwordMarker.JOINER)
        self.assertEqual(out, true_out)

    def test_subword_group_spacer(self):
        data_in = [
            "however",
            ",",
            "▁according",
            "▁to",
            "▁the",
            "▁logs",
            ",",
            "▁she",
            "▁is",
            "▁hard",
            "-",
            "working",
            ".",
        ]  # noqa: E501
        true_out = [0, 0, 1, 2, 3, 4, 4, 5, 6, 7, 7, 7, 7]
        out = subword_map_by_spacer(data_in, marker=SubwordMarker.SPACER)
        self.assertEqual(out, true_out)
        # no dummy prefix
        no_dummy = [
            "however",
            ",",
            "▁according",
            "▁to",
            "▁the",
            "▁logs",
            ",",
            "▁she",
            "▁is",
            "▁hard",
            "-",
            "working",
            ".",
        ]  # noqa: E501
        no_dummy_out = subword_map_by_spacer(no_dummy, marker=SubwordMarker.SPACER)
        self.assertEqual(no_dummy_out, true_out)

    def test_subword_group_spacer_with_case_markup(self):
        data_in = [
            "｟mrk_case_modifier_C｠",
            "▁however",
            ",",
            "▁according",
            "▁to",
            "▁the",
            "▁logs",
            ",",
            "▁｟mrk_begin_case_region_U｠",
            "▁she",
            "▁is",
            "▁hard",
            "-",
            "working",
            ".",
            "▁｟mrk_end_case_region_U｠",
        ]  # noqa: E501
        true_out = [0, 0, 0, 1, 2, 3, 4, 4, 5, 5, 6, 7, 7, 7, 7, 7]
        out = subword_map_by_spacer(data_in, marker=SubwordMarker.SPACER)
        self.assertEqual(out, true_out)

    def test_subword_group_spacer_with_spacer_new(self):
        data_in = [
            "｟mrk_case_modifier_C｠",
            "▁",
            "however",
            ",",
            "▁",
            "according",
            "▁",
            "to",
            "▁",
            "the",
            "▁",
            "logs",
            ",",
            "▁",
            "｟mrk_begin_case_region_U｠",
            "▁",
            "she",
            "▁",
            "is",
            "▁",
            "hard",
            "-",
            "working",
            ".",
            "▁",
            "｟mrk_end_case_region_U｠",
        ]  # noqa: E501
        true_out = [
            0,
            0,
            0,
            0,
            1,
            1,
            2,
            2,
            3,
            3,
            4,
            4,
            4,
            5,
            5,
            5,
            5,
            6,
            6,
            7,
            7,
            7,
            7,
            7,
            7,
            7,
        ]  # noqa: E501
        out = subword_map_by_spacer(data_in, marker=SubwordMarker.SPACER)
        self.assertEqual(out, true_out)


if __name__ == "__main__":
    unittest.main()



"""Base Scorer class and relate utils."""


class Scorer(object):
    """A Base class that every scoring method should derived from."""

    def __init__(self, config):
        # not used by any scorer for now it seems, but why not
        self.config = config

    def compute_score(self, preds, texts_refs):
        raise NotImplementedError


def build_scorers(config, scorers_cls):
    """Build scorers in `scorers_cls`."""
    scorers = {}
    for metric, scorer_cls in scorers_cls.items():
        scorer_obj = scorer_cls(config)
        scorers[metric] = {"scorer": scorer_obj, "value": 0}
    return scorers



from pydantic import BaseModel, ConfigDict  # , Enum


def get_config_dict():
    return ConfigDict(
        validate_assignment=True,
        validate_default=True,
        use_enum_values=True,
        extra="forbid",
        protected_namespaces=(),  # prevents warning for model_type
    )


class Config(BaseModel):
    """
    All config classes will inherit from this.
    Allows to homogenize model_config settings across all classes,
    and clarifies a bit naming standards (pydantic model_config vs our own).
    """

    model_config = get_config_dict()

    def update(self, **kwargs):
        updated_instance = self.copy(update=kwargs)
        self.__dict__.update(updated_instance.__dict__)
        self.__fields_set__.update(updated_instance.__fields_set__)



from tensorboard.backend.event_processing import event_accumulator
from argparse import ArgumentParser
import os


class TestEvents:
    def __init__(self):
        stats = ["xent", "ppl", "accuracy", "tgtper", "lr"]
        metrics = ["BLEU", "TER"]
        self.scalars = {}
        self.scalars["train"] = [("progress/" + stat) for stat in stats]
        self.scalars["valid"] = [("valid/" + stat) for stat in stats]
        self.scalars["valid_metrics"] = self.scalars["valid"] + [("valid/" + metric) for metric in metrics]

    def reload_events(self, path):
        ea = event_accumulator.EventAccumulator(
            path,
            size_guidance={event_accumulator.SCALARS: 0},
        )
        ea.Reload()
        return ea

    def check_scalars(self, scalars, logdir):
        for event_file in os.listdir(logdir):
            path = os.path.join(logdir, event_file)
            event_accumulator = self.reload_events(path)
            # make sure the scalars are in the event accumulator tags
            assert all(
                s in event_accumulator.Tags()["scalars"] for s in scalars
            ), "{} some scalars were not found in the event accumulator"


if __name__ == "__main__":
    # required arguments
    parser = ArgumentParser()
    requiredArgs = parser.add_argument_group("required arguments")
    requiredArgs.add_argument("-logdir", "--logdir", type=str, required=True)
    requiredArgs.add_argument(
        "-tensorboard_checks",
        "--tensorboard_checks",
        type=str,
        required=True,
        choices=["train", "valid", "valid_metrics"],
    )
    args = parser.parse_args()
    test_event = TestEvents()
    scalars = test_event.scalars[args.tensorboard_checks]
    print("looking for scalars: ", scalars, end=" ")
    test_event.check_scalars(scalars, args.logdir)



from eole.utils.logging import logger
from eole.transforms import register_transform
from .transform import Transform, TransformConfig
from pydantic import Field
import re


class TerminologyConfig(TransformConfig):
    termbase_path: str | None = Field(default=None, description="Path to a dictionary file with terms.")
    src_spacy_language_model: str | None = Field(
        default=None,
        description="Name of the spaCy language model for the source corpus.",
    )
    tgt_spacy_language_model: str | None = Field(
        default=None,
        description="Name of the spaCy language model for the target corpus.",
    )
    term_corpus_ratio: float | None = Field(default=0.3, description="Ratio of corpus to augment with terms.")
    term_example_ratio: float | None = Field(default=0.2, description="Maximum terms allowed in an example.")
    src_term_stoken: str | None = Field(default="｟src_term_start｠", description="The source term start token.")
    tgt_term_stoken: str | None = Field(default="｟tgt_term_start｠", description="The target term start token.")
    tgt_term_etoken: str | None = Field(default="｟tgt_term_end｠", description="The target term end token.")
    term_source_delimiter: str | None = Field(
        default="｟fuzzy｠",
        description="Any special token used for augmented source sentences. "
        "The default is the fuzzy token used in the "
        "FuzzyMatch transform.",
    )


class TermMatcher(object):
    def __init__(
        self,
        termbase_path,
        src_spacy_language_model,
        tgt_spacy_language_model,
        term_example_ratio,
        src_term_stoken,
        tgt_term_stoken,
        tgt_term_etoken,
        delimiter,
        term_corpus_ratio=0.2,
    ):
        import spacy

        self.term_example_ratio = term_example_ratio
        self.src_nlp = spacy.load(src_spacy_language_model, disable=["parser", "ner"])
        self.tgt_nlp = spacy.load(tgt_spacy_language_model, disable=["parser", "ner"])

        # We exclude tokenization for contractions in
        # order to avoid inconsistencies with pyeoleok's tokenization.
        # (e.g. "I ca n't" with spacy, "I can ' t" with pyonmttok)
        self.src_nlp.tokenizer.rules = {
            key: value
            for key, value in self.src_nlp.tokenizer.rules.items()
            if "'" not in key and "’" not in key and "‘" not in key
        }
        self.tgt_nlp.tokenizer.rules = {
            key: value
            for key, value in self.tgt_nlp.tokenizer.rules.items()
            if "'" not in key and "’" not in key and "‘" not in key
        }
        self.internal_termbase = self._create_internal_termbase(termbase_path)
        self.automaton = self._create_automaton()
        self.term_corpus_ratio = term_corpus_ratio
        self.src_term_stoken = src_term_stoken
        self.tgt_term_stoken = tgt_term_stoken
        self.tgt_term_etoken = tgt_term_etoken
        self.delimiter = delimiter

    def _create_internal_termbase(self, termbase_path):
        logger.debug("Creating termbase with lemmas for Terminology transform")

        # Use Spacy's stopwords to get rid of junk entries
        src_stopwords = self.src_nlp.Defaults.stop_words
        tgt_stopwords = self.tgt_nlp.Defaults.stop_words
        termbase = list()
        with open(termbase_path, mode="r", encoding="utf-8") as file:
            pairs = file.readlines()
            for pair in pairs:
                src_term, tgt_term = map(str, pair.split("\t"))
                src_lemma = " ".join("∥".join(tok.lemma_.split(" ")) for tok in self.src_nlp(src_term)).strip()
                tgt_lemma = " ".join(tok.lemma_ for tok in self.tgt_nlp(tgt_term)).strip()
                if src_lemma.lower() not in src_stopwords and tgt_lemma.lower() not in tgt_stopwords:
                    termbase.append((src_lemma, tgt_lemma))
        logger.debug(f"Created termbase with {len(termbase)} lemmas " f"for Terminology transform")
        return termbase

    def _create_automaton(self):
        import ahocorasick

        automaton = ahocorasick.Automaton()
        for term in self.internal_termbase:
            automaton.add_word(term[0], (term[0], term[1]))
        automaton.make_automaton()
        return automaton

    def _src_sentence_with_terms(self, source_string, target_string) -> tuple:

        maybe_augmented = source_string.split(self.delimiter)
        source_only = maybe_augmented[0].strip()
        augmented_part = maybe_augmented[1].strip() if len(maybe_augmented) > 1 else None

        doc_src = self.src_nlp(source_only)
        doc_tgt = self.tgt_nlp(target_string)

        # Perform tokenization with spacy for consistency.
        tokenized_source = [tok.text for tok in doc_src]
        lemmatized_source = ["∥".join(tok.lemma_.lower().split(" ")) for tok in doc_src]
        lemmatized_target = [tok.lemma_.lower() for tok in doc_tgt]

        lemmatized_source_string = " ".join(lemmatized_source)

        offset = 0
        source_with_terms = list()
        term_counter = 0

        max_terms_allowed = int(len(tokenized_source) * self.term_example_ratio)
        is_match = False
        for match_end, (src_entry, tgt_entry) in self.automaton.iter_long(lemmatized_source_string):

            if term_counter == max_terms_allowed:
                break

            match_start = match_end - len(src_entry) + 1

            # We ensure that the target lemma is present in the lemmatized
            # target string, that the match is an exact match (there is
            # whitespace before or after the term)
            # and we perform some bound checking.
            if (
                (tgt_entry.lower() not in " ".join(lemmatized_target).lower())
                or (
                    len(lemmatized_source_string) != match_end + 1
                    and not (lemmatized_source_string[match_end + 1].isspace())
                )
                or (not lemmatized_source_string[match_start - 1].isspace() and match_start != 0)
            ):
                continue
            else:
                term_counter += 1

                # Map the lemmatized string match index to
                # the lemmatized list index
                lemma_list_index = 0
                for i, w in enumerate(lemmatized_source):
                    if lemma_list_index == match_start:
                        lemma_list_index = i
                        break
                    else:
                        lemma_list_index += len(w) + 1

                # We need to know if the term is multiword
                num_words_in_src_term = len(src_entry.split(" "))
                src_term = " ".join(
                    tokenized_source[lemma_list_index : lemma_list_index + num_words_in_src_term]
                ).strip()

                # Join multiword target lemmas with a unique separator so
                # we can treat them as single word and not change the indices.
                tgt_term = tgt_entry.replace(" ", "∥").rstrip().lower()
                source_with_terms.append(
                    f"{lemmatized_source_string[offset: match_start]}"
                    f"{self.src_term_stoken}∥{src_term}∥{self.tgt_term_stoken}∥"
                    f"{tgt_term}∥{self.tgt_term_etoken}"
                )

                offset = match_end + 1
                is_match = True

        if is_match:
            source_with_terms.append(lemmatized_source_string[offset:])
            tokenized_source_with_terms = "".join(source_with_terms).split(" ")

            if not (len(tokenized_source) == len(lemmatized_source) == len(tokenized_source_with_terms)):
                final_string = " ".join(tokenized_source)
                fixed_punct = re.sub(r" ([^\w\s｟\-\–])", r"\1", final_string)
                return fixed_punct.split(" "), not is_match

            # Construct the final source from the lemmatized list
            # that contains the terms. We compare the tokens in the
            # term-augmented lemma list with the tokens in the original
            # lemma list. If the lemma is the same, then we replace with
            # the token from the original tokenized source list. If they
            # are not the same, it means the lemma has been augemented
            # with a term, so we inject this in the final list.
            completed_tokenized_source = list()
            for idx in range(len(tokenized_source_with_terms)):
                # Restore the spaces in multi-word terms
                src_lemma = tokenized_source_with_terms[idx].replace("∥", " ")
                if lemmatized_source[idx].replace("∥", " ") == src_lemma:
                    completed_tokenized_source.append(tokenized_source[idx])
                else:
                    completed_tokenized_source.append(src_lemma)

            if augmented_part is not None:
                final_string = " ".join(completed_tokenized_source + [self.delimiter] + augmented_part.split(" "))
            else:
                final_string = " ".join(completed_tokenized_source)

            fixed_punct = re.sub(r" ([^\w\s｟\-\–])", r"\1", final_string)
            return fixed_punct.split(" "), is_match
        else:
            final_string = " ".join(tokenized_source)
            fixed_punct = re.sub(r" ([^\w\s｟\-\–])", r"\1", final_string)
            return fixed_punct.split(" "), not is_match


@register_transform(name="terminology")
class TerminologyTransform(Transform):

    config_model = TerminologyConfig

    def __init__(self, config):
        super().__init__(config)

    def _parse_config(self):
        self.termbase_path = self.config.termbase_path
        self.src_spacy_language_model = self.config.src_spacy_language_model
        self.tgt_spacy_language_model = self.config.tgt_spacy_language_model
        self.term_corpus_ratio = self.config.term_corpus_ratio
        self.term_example_ratio = self.config.term_example_ratio
        self.term_source_delimiter = self.config.term_source_delimiter
        self.src_term_stoken = self.config.src_term_stoken
        self.tgt_term_stoken = self.config.tgt_term_stoken
        self.tgt_term_etoken = self.config.tgt_term_etoken

    @classmethod
    def get_specials(cls, config):
        """Add the term tokens to the src vocab."""
        src_specials = list()
        src_specials.extend([config.src_term_stoken, config.tgt_term_stoken, config.tgt_term_etoken])
        return (src_specials, list())

    def warm_up(self, vocabs=None):
        """Create the terminology matcher."""

        super().warm_up(None)
        self.termmatcher = TermMatcher(
            self.termbase_path,
            self.src_spacy_language_model,
            self.tgt_spacy_language_model,
            self.term_example_ratio,
            self.src_term_stoken,
            self.tgt_term_stoken,
            self.tgt_term_etoken,
            self.term_source_delimiter,
            self.term_corpus_ratio,
        )

    def batch_apply(self, batch, is_train=False, stats=None, **kwargs):
        bucket_size = len(batch)
        examples_with_terms = 0

        for i, (ex, _, _) in enumerate(batch):
            # Skip half examples to improve performance. This means we set
            # a hard limit for the `term_corpus_ratio` to 0.5, which is actually
            # quite high. TODO: We can add this (skipping examples) as an option
            if i % 2 == 0:
                original_src = ex["src"]
                augmented_example, is_match = self.apply(ex, is_train, stats, **kwargs)
                if is_match and (examples_with_terms < bucket_size * self.term_corpus_ratio):
                    examples_with_terms += 1
                    ex["src"] = augmented_example["src"]
                else:
                    ex["src"] = original_src

        logger.debug(f"Added terms to {examples_with_terms}/{bucket_size} examples")
        return batch

    def apply(self, example, is_train=False, stats=None, **kwargs) -> tuple:
        """Add terms to source examples."""

        example["src"], is_match = self.termmatcher._src_sentence_with_terms(
            " ".join(example["src"]), " ".join(example["tgt"])
        )
        return example, is_match



import torch
from typing import List, Literal, Union
from pydantic import Field, field_validator
from importlib import import_module
from eole.config.config import Config, get_config_dict
from eole.constants import TORCH_DTYPES

# from eole.utils.logging import logger


class DistributedConfig(Config):
    gpu_ranks: List[int] = Field(default=[], description="List of ranks for each process.")
    world_size: int = Field(default=1, description="Total number of distributed processes.")
    parallel_mode: Literal["data_parallel", "tensor_parallel"] = Field(
        default="data_parallel", description="Distributed mode."
    )
    gpu_backend: str = Field(default="nccl", description="Type of torch distributed backend.")
    gpu_verbose_level: int = Field(default=0, description="Gives more info on each process per GPU.")
    master_ip: str = Field(default="localhost", description="IP of master for torch.distributed training.")
    master_port: int = Field(default=10000, description="Port of master for torch.distributed training.")
    timeout: int = Field(default=60, description="Timeout for one GPU to wait for the others.")

    @property
    def parallel_gpu(self) -> int:  # converted to a `property` by `computed_field`
        return self.world_size if self.parallel_mode == "tensor_parallel" else 1

    # This could be a good idea, but requires more safeguards, e.g. for inference
    # @model_validator(mode="after")
    # def _validate_distributed_config(self):
    #     if torch.cuda.is_available() and "gpu_ranks" not in self.model_fields_set:
    #         logger.warning(
    #             "GPUs detected, using them all. "
    #             "Explicitly set gpu_ranks/world_size to change this behaviour.")
    #         num_gpus = torch.cuda.device_count()
    #         self.world_size = num_gpus
    #         self.gpu_ranks = list(range(num_gpus))
    #     return self


class LoRaConfig(Config):
    lora_layers: List[str] = Field(
        default=[],
        description="List of layers to be replaced by LoRa layers. "
        "E.g. ['linear_values', 'linear_query'] "
        "(§4.2 in https://arxiv.org/abs/2106.09685)",
    )  # validate against existing layers?
    lora_embedding: bool = Field(default=False, description="Replace embeddings with LoRa Embeddings (§5.1)")
    lora_rank: int = Field(default=2, description="r=2 successfully tested with NLLB-200 3.3B")
    lora_alpha: int = Field(default=1, description="§4.1 https://arxiv.org/abs/2106.09685")
    lora_dropout: float = Field(
        default=0.0, description="Rule of thumb: same value as in main model."
    )  # should probably be in TrainingConfig in fine to be coherent with dropout/attention_dropout


class QuantizeConfig(Config):
    # Quantization related options (previously in MiscConfig)
    quant_layers: List[str] = Field(
        default=[], description="List of layers to be compressed in 4/8bit."
    )  # validate against list of layers names ?
    quant_type: Literal["", "bnb_8bit", "bnb_FP4", "bnb_NF4", "awq_gemm", "awq_gemv"] = Field(
        default="", description="Type of compression."
    )
    w_bit: int = Field(default=4, description="W_bit quantization")  # single authorized value for now actually
    group_size: int = Field(default=128, description="Group size quantization.")  # same


class MiscConfig(Config):
    # Some opts not super sure where to put
    # Seed stuff
    seed: int = Field(
        default=-1,
        description="Set random seed used for better reproducibility between experiments.",
    )


class LoggingConfig(Config):
    log_file: str = Field(default="", description="Output logs to a file under this path.")
    # some "observability" stuff
    report_every: int = Field(default=50, description="Print stats at this interval (in steps).")
    valid_metrics: List[str] = Field(
        default=[], description="List of names of additional validation metrics."
    )  # should probably be validated properly (or in some registry/enum)
    scoring_debug: bool = Field(default=False, description="Dump src/ref/pred of the current batch.")
    dump_preds: str | None = Field(default=None, description="Folder to dump predictions to.")
    tensorboard: bool = Field(default=False, description="Use tensorboard for visualization during training.")
    tensorboard_log_dir: str = Field(
        default="runs/eole",
        description="Log directory for tensorboard (also the name of the run).",
    )
    tensorboard_log_dir_dated: str | None = (
        None  # field dynamically populated in build_report_manager, might improve later
    )


class RunningConfig(DistributedConfig):
    """
    Should be used as base class for Training/Inference at some point,
    factorizing common stuff like batch_size etc.
    """

    model_config = get_config_dict()
    model_config["arbitrary_types_allowed"] = True  # to allow torch.dtype

    model_path: str = Field(
        default="model",
        description="Path to directory containing all model components.",
    )
    self_attn_backend: Literal["flash", "pytorch"] = Field(
        default="flash",
        description="Self-attention backend.",
    )
    compute_dtype: Union[Literal["fp32", "fp16", "int8", "bf16"], torch.dtype] = Field(
        default=torch.float32,
        description="Compute dtype (precision) to use for main compute. "
        "Some parameters might have other dtypes for specific cases "
        "(e.g. torch.amp -- See eole.config.training.TrainingConfig.storage_dtype) "
        "fp32 to force slow fp16 model on gtx1080, "
        "int8 to enable pytorch native 8-bit quantization (cpu only).",
    )
    torch_compile: bool = Field(default=False, description="Use torch.compile with dynamic=True.")

    @field_validator("compute_dtype", mode="before")
    @classmethod
    def validate_compute_dtype(cls, v: Union[str, torch.dtype]) -> torch.dtype:
        if isinstance(v, str):
            if v in TORCH_DTYPES:
                return TORCH_DTYPES[v]
            else:
                raise ValueError(f"Invalid compute_dtype value: {v}")
        return v

    def check_self_attn_backend(self):
        try:
            flash_pack = import_module("flash_attn")
            if (
                hasattr(flash_pack, "flash_attn_func")
                and torch.cuda.get_device_capability()[0] >= 8
                and self.self_attn_backend == "flash"
                and self.compute_dtype in [torch.float16, torch.bfloat16]
            ):
                pass
            else:
                self.__dict__["self_attn_backend"] = "pytorch"
        except ImportError:
            self.__dict__["self_attn_backend"] = "pytorch"



#!/usr/bin/env python
# Standard Library Imports
import configparser
import json
import logging
import math
import os
import re
from dataclasses import dataclass, field, fields
from typing import Optional

# Third-Party Library Imports
import huggingface_hub
import pyonmttok
import safetensors
import torch
from huggingface_hub import hf_hub_download, utils
from safetensors.torch import save_file
from sentencepiece import SentencePieceProcessor

# Eole Imports
from eole.bin import BaseBin, register_bin
from eole.config import recursive_model_fields_set
from eole.config.run import TrainConfig
from eole.config.training import TrainingConfig
from eole.constants import DefaultTokens, TORCH_DTYPES, PositionEncodingType
from eole.inputters.inputter import vocabs_to_dict

from eole.bin.convert.HF_mappings import (
    KEY_MAPS,
    ACT_TABLE,
    LN_TABLE,
    ARCH_TABLE,
    TOK_TABLE,
)


def get_sentencepiece_vocab(model_path):
    """Get the vocabulary from a SentencePiece model.

    Args:
        model_path (str): Path to the SentencePiece model file

    Returns:
        list: The vocabulary as a list of strings
    """
    sp_model = SentencePieceProcessor(model_file=model_path)
    n_words = sp_model.vocab_size()
    return [sp_model.id_to_piece(i) for i in range(n_words)]


@dataclass
class HuggingfaceFiles:
    """
    Helper class to centralize HF files and config processing.
    """

    model_dir: str
    token: str
    # files retrieved from the hub
    config_path: str
    tokenizer_config_json: str
    generation_config_json: Optional[str] = None
    tokenizer_model: Optional[str] = None
    tokenizer_json: Optional[str] = None
    wmap_path: Optional[str] = None
    model_path: Optional[str] = None
    special_tokens_json: Optional[str] = None

    # Unified dictionary to cache loaded files
    _loaded_files: dict = field(default_factory=dict, init=False)

    @classmethod
    def fetch(cls, args, mode="local"):
        """
        Factory method to fetch files either locally or from the HuggingFace Hub.

        Args:
            args: Namespace containing model_dir, token, output, etc.
            mode: Either 'local' or 'hub', determines the source of the files.

        Returns:
            HuggingfaceFiles instance with all paths populated.
        """

        if os.path.exists(args.model_dir):
            mode = "local"
        else:
            mode = "hub"

        def get_local_path(file_name, required=False):
            """Helper to check if a file exists in the model directory."""
            path = os.path.join(args.model_dir, file_name)
            if os.path.exists(path):
                return path
            if required:
                raise FileNotFoundError(f"Required file '{file_name}' is missing locally.")
            return None

        def download_file_from_hub(file_name, required=True):
            """Helper to download a file from the HuggingFace Hub."""
            try:
                return hf_hub_download(
                    repo_id=args.model_dir,
                    filename=file_name,
                    token=args.token,
                    local_dir=args.output,
                )
            except utils.EntryNotFoundError:
                if required:
                    raise utils.EntryNotFoundError(f"Required file '{file_name}' is missing from the repository.")
                return None

        get_file_fn = get_local_path if mode == "local" else download_file_from_hub

        os.makedirs(args.output, exist_ok=True)  # Ensure output directory exists

        # Fetch required and optional files
        paths = {
            "config_path": get_file_fn("config.json", required=True),
            "tokenizer_config_json": get_file_fn("tokenizer_config.json", required=True),
            "generation_config_json": get_file_fn("generation_config.json", required=False),
            "tokenizer_model": get_file_fn("tokenizer.model", required=False)
            or get_file_fn("sentencepiece.bpe.model", required=False),
            "tokenizer_json": get_file_fn("tokenizer.json", required=False),
            "wmap_path": get_file_fn("model.safetensors.index.json", required=False)
            or get_file_fn("pytorch_model.bin.index.json", required=False),
            "model_path": get_file_fn("model.safetensors", required=False)
            or get_file_fn("pytorch_model.bin", required=False),
            "special_tokens_json": get_file_fn("special_tokens_map.json", required=False),
        }

        return cls(**paths, model_dir=args.model_dir, token=args.token)

    def _get_file_content(self, file_attr):
        """Combined method to load and cache file contents."""
        file_path = getattr(self, file_attr, None)
        if not file_path or not os.path.exists(file_path):
            return None

        if file_path in self._loaded_files:
            return self._loaded_files[file_path]

        with open(file_path, encoding="utf-8") as f:
            content = json.load(f) if file_path.endswith(".json") else f.read()

        self._loaded_files[file_path] = content
        return content

    def __getattr__(self, name):
        """Generic attribute getter for file contents."""
        file_fields = {field.name for field in fields(self)}
        if f"{name}_json" in file_fields:
            return self._get_file_content(f"{name}_json")
        elif f"{name}_path" in file_fields:
            return self._get_file_content(f"{name}_path")
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")

    @property
    def arch(self):
        return self.config["architectures"][0]

    @property
    def vocab_size(self):
        config = self.config.get("text_config", self.config)
        if "vocab_size" in config:
            return config["vocab_size"]
        tokenizer = self.tokenizer
        if tokenizer:
            max_added_tokens = max([token["id"] for token in tokenizer["added_tokens"]])
            max_tokens = max([token_id for token_id in tokenizer["model"]["vocab"].values()])
            return max(max_added_tokens, max_tokens)
        raise NotImplementedError("Vocabulary size could not be determined.")

    @property
    def encoder_layer_prefix(self):
        return KEY_MAPS[self.arch].get("encoder_layer_prefix", None)

    @property
    def encoder_sam_layer_prefix(self):
        return KEY_MAPS[self.arch].get("encoder_sam_layer_prefix", None)

    @property
    def decoder_layer_prefix(self):
        return KEY_MAPS[self.arch].get("decoder_layer_prefix", None)

    @property
    def base_dir(self):
        return os.path.split(self.wmap_path)[0]

    def get_load_ckpt(self, dir_path, file_path):
        if os.path.exists(os.path.join(dir_path, file_path)):
            ckpt_path = os.path.join(dir_path, file_path)
        else:
            try:
                ckpt_path = huggingface_hub.hf_hub_download(
                    repo_id=self.model_dir,
                    filename=file_path,
                    token=self.token,
                )
            except huggingface_hub.utils.EntryNotFoundError:
                raise huggingface_hub.utils.EntryNotFoundError("Checkpoint not found on the hub")
            except PermissionError:
                ckpt_path = os.path.join(dir_path, file_path)
        if ckpt_path[-4:] == ".bin":
            checkpoint = torch.load(ckpt_path, map_location=torch.device("cpu"))
        else:
            checkpoint = ckpt_path

        return checkpoint

    def checkpoint(self, ckpt):
        if self.wmap_path:
            checkpoint = self.get_load_ckpt(self.base_dir, ckpt)
        else:
            checkpoint = self.get_load_ckpt(*os.path.split(self.model_path))
        return checkpoint


def get_git_remote_url(model_dir):
    """Gets the git remote URL from a model directory.

    Args:
        model_dir (str): Path to model directory containing .git folder

    Returns:
        str: Repository name in format "owner/repo", or None if not found
    """
    config_path = os.path.join(model_dir, ".git", "config")
    config = configparser.ConfigParser()
    config.read(config_path)
    # Get the URL from the 'remote "origin"' section
    try:
        url = config['remote "origin"']["url"]
        return "/".join(url.split("/")[-2:])
    except KeyError:
        return None


def get_huggingface_model(args):
    """Gets HuggingFace model identifier from directory path or direct model name.

    Args:
        args: Arguments containing model_dir attribute with path to model directory or model identifier

    Returns:
        str: HuggingFace model identifier
    """
    if os.path.exists(args.model_dir):
        huggingface_model = get_git_remote_url(args.model_dir)
        if huggingface_model is None:
            logging.warning("huggingface_model could not be retrieved from model_dir git config")
    else:
        huggingface_model = args.model_dir
    return huggingface_model


def build_config_dict(hf):
    """Convert Hugging Face model configuration to eole config format.

    This function takes a HuggingfaceFiles dataclass object and extracts configuration parameters
    to create dictionaries for model architecture, training settings and various hyper-parameters.

    Args:
        hf: HuggingfaceFiles dataclass object containing config and architecture information

    Returns:
        tuple: Contains:
            - model_config (dict): Model architecture configuration
            - training_config (dict): Training and quantization settings
            - params (list): List of parameter names to extract
    """
    config = hf.config
    arch = hf.arch
    print("Architecture: ", arch)

    vision_config = config.get("vision_config", None)
    other_config = config  # save what is not text/vision for later use
    config = config.get("text_config", config.get("language_config", config))

    model_config = {}
    training_config = {}

    # Initialize model_config with defaults and fallbacks
    model_config = {
        "layers": config.get("num_hidden_layers", config.get("n_layer", config.get("n_layers"))),
        "hidden_size": config.get("hidden_size", config.get("n_embd", config.get("hidden_dim", config.get("d_model")))),
        "heads": config.get(
            "num_attention_heads",
            config.get("n_head", config.get("n_heads", config.get("decoder_attention_heads", 32))),
        ),  # default 32 patch for mistral-community/pixtral-12b
        "transformer_ff": config.get("intermediate_size", config.get("decoder_ffn_dim", None)),
        "mlp_activation_fn": ACT_TABLE[arch],
        "layer_norm": LN_TABLE[arch],
        "heads_kv": config.get("multi_query", False)
        or config.get(
            "num_key_value_heads",
            config.get(
                "num_kv_heads",
                config.get("n_head_kv", config.get("num_attention_heads", config.get("n_head"))),
            ),
        ),
        "head_dim": config.get("head_dim"),
        "attn_scaling": config.get("query_pre_attn_scalar", None),
        "parallel_residual": config.get("parallel_attn", False),
        "norm_eps": config.get(
            "rms_norm_eps",
            config.get("layer_norm_epsilon", config.get("layer_norm_eps", 1e-5)),
        ),
        "sliding_window": config.get("sliding_window", 0) or 4096,
        "num_experts": config.get("num_local_experts", config.get("n_routed_experts", config.get("num_experts", 0))),
        "num_shared_experts": config.get("n_shared_experts", 0),
        "first_k_dense_replace": config.get("first_k_dense_replace", 0),
        "num_experts_per_tok": config.get("num_experts_per_tok", 0),
        "moe_transformer_ff": config.get("moe_intermediate_size", None),
        "add_qkvbias": False,
        "add_final_linear_bias": False,
        "add_ffnbias": False,
        "shared_layer_norm": False,
        "left_pad": True,
        "generator_bias": False,
        "adapter_bias": False,
        "rope_config": {
            "rotary_interleave": False,
            "rotary_theta": config.get("rope_theta", 10000),
        },
        "embeddings": {},  # Populated later
    }
    if model_config["num_experts"] == 1:
        model_config["num_experts"] = 0

    # Vision encoder
    if vision_config is not None:
        model_config["encoder"] = {
            "layers": vision_config.get("num_hidden_layers", 24),  # hard-coded for mistral-community/pixtral-12b
            "hidden_size": vision_config.get("hidden_size", vision_config.get("image_size", 0)),
            "transformer_ff": vision_config.get("intermediate_size", vision_config.get("image_size", 0) * 4),
            "image_size": vision_config.get("image_size", 0),
            "patch_size": vision_config.get("patch_size", 0),
            "heads": vision_config.get(
                "num_attention_heads", vision_config.get("image_size", 0) / vision_config.get("head_dim", 1)
            ),
            "heads_kv": vision_config.get(
                "num_attention_heads", vision_config.get("image_size", 0) / vision_config.get("head_dim", 1)
            ),
            "head_dim": vision_config.get(
                "head_dim", vision_config.get("hidden_size", 0) // vision_config.get("num_attention_heads", 1)
            ),
        }
    if arch in ["LlavaForConditionalGeneration", "Mistral3ForConditionalGeneration"]:
        model_config["encoder"].update(
            {
                "mlp_activation_fn": model_config["mlp_activation_fn"],
                "layer_norm": model_config["layer_norm"],
                "norm_eps": model_config["norm_eps"],
                "rope_config": {
                    "rotary_theta": vision_config["rope_theta"],
                    "rotary_interleave": False,
                },
            }
        )
        model_config["adapter_bias"] = other_config.get("multimodal_projector_bias", False)
        model_config["projector_activation_fn"] = other_config.get("projector_hidden_act", "gelu")
        model_config["spatial_merge_size"] = other_config.get("spatial_merge_size", None)

    if arch == "Gemma3ForConditionalGeneration":
        if model_config.get("head_dim", None) is None:
            model_config["head_dim"] = 256  # src/transformers/models/gemma3/configuration_gemma3.py#L61
        if model_config.get("heads_kv", None) is None:
            model_config["heads_kv"] = 4
        if model_config.get("heads", None) is None:
            model_config["heads"] = 8
        model_config["encoder"].update(
            {
                "n_positions": (vision_config["image_size"] // vision_config["patch_size"]) ** 2,
                # head related stuff patched to match 1152 dim of siglip
                # https://github.com/huggingface/transformers/blob/main/src/transformers/models/siglip/modeling_siglip.py#L399-L402
                "mm_tokens_per_image": hf.config["mm_tokens_per_image"],
                "image_token_id": hf.config["image_token_index"],
            }
        )

    if arch in ["DeepseekOCRForCausalLM"]:
        model_config["encoder"].update(
            {
                "n_positions": (
                    vision_config["width"]["clip-l-14-224"]["image_size"]
                    // vision_config["width"]["clip-l-14-224"]["patch_size"]
                )
                ** 2
                + 1,
                "patch_size": vision_config["width"]["clip-l-14-224"].get("patch_size", 14),
                "heads": vision_config["width"]["clip-l-14-224"].get("heads"),
                "heads_kv": vision_config["width"]["clip-l-14-224"].get("heads"),
            }
        )

    if arch in ["HunYuanVLForConditionalGeneration"]:
        model_config["encoder"].update({})
        model_config["spatial_merge_size"] = vision_config.get("spatial_merge_size", None)

    # patch transformer_ff
    if model_config["transformer_ff"] is None:
        model_config["transformer_ff"] = model_config["hidden_size"] * 4

    # patch sliding window
    if model_config["sliding_window"] is None:
        model_config["sliding_window"] = 4096

    # Populate embeddings
    model_config["embeddings"] = {
        "src_word_vec_size": model_config["hidden_size"],
        "tgt_word_vec_size": model_config["hidden_size"],
    }

    # Default position encoding configuration
    model_config["embeddings"].update(
        {
            "position_encoding_type": PositionEncodingType.Rotary,
            "n_positions": 0,
        }
    )

    # patch rotary dim
    if "rotary_dim" in config.keys():
        model_config["rope_config"]["rotary_dim"] = config["rotary_dim"]
    elif "partial_rotary_factor" in config.keys():
        model_config["rope_config"]["rotary_dim"] = int(
            config["partial_rotary_factor"] * (model_config["hidden_size"] // model_config["heads"])
        )
    elif model_config.get("head_dim", None) is not None:
        model_config["rope_config"]["rotary_dim"] = model_config["head_dim"]
    else:
        model_config["rope_config"]["rotary_dim"] = model_config["hidden_size"] // model_config["heads"]

    # Update rope scaling related settings
    if config.get("rope_scaling", None) is not None:
        model_config["rope_config"].update(
            {
                "scaling_type": config["rope_scaling"].get("rope_type", config["rope_scaling"].get("type", None)),
                "scaling_factor": config["rope_scaling"].get("factor", 8.0),
                "alpha": config["rope_scaling"].get("alpha", None),
                "low_freq_factor": config["rope_scaling"].get("low_freq_factor", 1.0),
                "high_freq_factor": config["rope_scaling"].get("high_freq_factor", 4.0),
                "original_max_position_embeddings": config["rope_scaling"].get(
                    "original_max_position_embeddings", config.get("max_position_embeddings", 8192)
                ),
            }
        )

    # Validate required fields
    required_fields = {
        "layers": "Can't find the number of layers in the config.json file",
        "hidden_size": "Can't find the model hidden size in the config.json file",
        "heads": "Can't find the number of heads in the config.json file",
    }

    for key, error_msg in required_fields.items():
        if model_config[key] is None:
            raise ValueError(error_msg)

    # Handle quantization
    quant_config = config.get("quantization_config", {})
    if quant_config:
        if quant_config.get("quant_method") == "awq":
            backend = quant_config.get("backend", "autoawq")
            quant_type_mapping = {
                "llm-awq": "awq_gemv",
                "autoawq": {"gemm": "awq_gemm", "gemv": "awq_gemv"},
            }
            quant_type = quant_type_mapping.get(backend, "").get(quant_config.get("version").lower(), "")
            if not quant_type:
                raise ValueError("Unknown quantization config")
        else:
            raise ValueError("Can convert only awq models for now")
        training_config["quant_type"] = quant_type
        training_config["w_bit"] = quant_config.get("bits", quant_config.get("w_bit", 0))
        training_config["group_size"] = quant_config.get("group_size", quant_config.get("q_group_size", 0))
        training_config["quant_layers"] = [
            "gate_up_proj",
            "down_proj",
            "up_proj",
            "linear_values",
            "linear_query",
            "linear_keys",
            "final_linear",
        ]
        params = ["qweight", "qzeros", "scales"] + ["weight", "bias"]
    else:
        training_config["quant_type"] = ""
        training_config["w_bit"] = 0
        training_config["group_size"] = 0
        training_config["quant_layers"] = []
        params = ["weight", "bias"]

    model_config["share_decoder_embeddings"] = config.get("tie_word_embeddings", False)

    # Update model_config based on architecture
    if arch in KEY_MAPS:
        for key, value in KEY_MAPS[arch].get("config", {}).items():
            if isinstance(value, dict):
                # Update nested dictionaries
                model_config[key] = {**model_config.get(key, {}), **value}
            else:
                # Update regular keys
                model_config[key] = value

    return model_config, training_config, params


def get_weight(checkpoint, tensor_name):
    if isinstance(checkpoint, dict):
        if tensor_name in checkpoint.keys():
            return checkpoint[tensor_name].contiguous()
    else:
        with safetensors.safe_open(checkpoint, framework="pt", device="cpu") as f:
            if tensor_name in f.keys():
                return f.get_tensor(tensor_name).contiguous()
    return None


def check_tokenizer_config(hf):
    config = hf.config
    add_bos_token = hf.tokenizer_config.get("add_bos_token", hf.tokenizer_config.get("bos_token", None) is not None)
    chat_template = {"chat_template": hf.tokenizer_config.get("chat_template", None)}
    eos_token_id = config.get("eos_token_id", None)
    optional_eos = []
    if isinstance(eos_token_id, list) and "added_tokens_decoder" in hf.tokenizer_config:
        eos_tokens = [hf.tokenizer_config["added_tokens_decoder"][str(index)]["content"] for index in eos_token_id]
        optional_eos = eos_tokens[1:]
    # Automatically convert added_tokens into mapped_tokens
    mapped_tokens = [
        (
            token["content"],
            re.sub(r"<\|([^|]*)\|>", "\uff5f\\1\uff60", token["content"]),
        )
        for token in hf.tokenizer_config.get("added_tokens_decoder", {}).values()
    ]
    return (add_bos_token, chat_template, optional_eos, mapped_tokens)


def check_special_tokens(hf):
    vocabs = {"specials": {}}
    if hf.special_tokens is not None:
        special_tokens_map = hf.special_tokens
        for token_name in ["bos_token", "unk_token", "eos_token", "pad_token"]:
            token = special_tokens_map.get(token_name, None)
            if isinstance(token, list):
                vocabs["specials"][token_name] = token[0]
            elif isinstance(token, str):
                vocabs["specials"][token_name] = token
            elif isinstance(token, dict):
                vocabs["specials"][token_name] = token["content"]
    elif hf.tokenizer_config is not None:
        for token_name in ["bos_token", "unk_token", "eos_token", "pad_token"]:
            token = hf.tokenizer_config.get(token_name, None)
            if isinstance(token, list):
                vocabs["specials"][token_name] = token[0]
            elif isinstance(token, str):
                vocabs["specials"][token_name] = token
            elif isinstance(token, dict):
                vocabs["specials"][token_name] = token["content"]
    return vocabs


def check_generation_config(hf):
    if hf.generation_config is None:
        return {}
    generation_config_dict = {}
    # we probably need a better mapping at some point
    keys = ["top_k", "top_p", "temperature", "max_length"]
    generation_config_dict = {key: hf.generation_config[key] for key in keys if key in hf.generation_config}
    return generation_config_dict


def get_shards_map(model_config, hf, nshards):
    """
    Distributes model checkpoints across shards based on layer ranges.

    Args:
        model_config (dict): Configuration dictionary containing model parameters,
            including number of layers.
        hf (object): HuggingfaceFiles dataclass object containing model path and weight mapping.
        nshards (int): Number of shards to distribute the model across.

    Returns:
        tuple: A tuple containing:
            - list of lists: Checkpoint paths for each shard
            - list of ranges: Layer ranges assigned to each shard
    """
    # Compute layer range for each shard
    if "encoder" in model_config.keys():
        n_layers = max(model_config["layers"], model_config["encoder"]["layers"])
    else:
        n_layers = model_config["layers"]
    layers_per_shard = math.ceil(n_layers / nshards)
    shard_layer_ranges = [
        range(
            layers_per_shard * shard,
            min(layers_per_shard * (shard + 1), n_layers),
        )
        for shard in range(nshards)
    ]
    if hf.wmap_path is None:
        return [[hf.model_path]] * nshards, shard_layer_ranges

    weightmap = hf.wmap["weight_map"]

    # Initialize a list of checkpoint lists, one for each shard
    shard_checkpoints = [set() for _ in range(nshards)]

    # Check if a layer key belongs to the current shard
    def is_layer_in_range(key, prefix, layer_range):
        return prefix and key.startswith(prefix) and int(key.split(".")[len(prefix.split(".")) - 1]) in layer_range

    # Loop over the weightmap and distribute checkpoints to the appropriate shards
    for key, ckpt in weightmap.items():
        for shard, layer_range in enumerate(shard_layer_ranges):
            if is_layer_in_range(key, hf.decoder_layer_prefix, layer_range) or is_layer_in_range(
                key, hf.encoder_layer_prefix, layer_range
            ):
                shard_checkpoints[shard].add(ckpt)

    return shard_checkpoints, shard_layer_ranges


def build_shards(model_config, hf, args, params):
    """
    Build sharded model files from HuggingFace checkpoint.

    Args:
        model_config (dict): Configuration dictionary containing model architecture settings
        hf (HuggingfaceFiles): dataclass containing model files and configuration
        args (Namespace): Command line arguments containing nshards and output path
        params (list): List of parameter names to convert

    The function splits the model into shards and saves them as safetensor files.
    Layer parameters are distributed across shards based on the sharding configuration.
    The first shard contains embeddings and model-level parameters on top of its layer split.
    """
    shard_checkpoints, shard_layer_ranges = get_shards_map(model_config, hf, args.nshards)

    for shard in range(args.nshards):

        print("starting output shard: %d/%d" % (shard + 1, args.nshards))
        eole_safetensor = {}

        def build_first_shard(hf, eole_safetensor):
            for target in KEY_MAPS[hf.arch].keys():
                if model_config["share_decoder_embeddings"] and target == "generator.weight":
                    continue
                source = KEY_MAPS[hf.arch][target]
                srckey, srcmap = source if isinstance(source, tuple) else (source, None)
                if isinstance(srckey, str):
                    if hf.wmap_path:
                        if srckey in hf.wmap["weight_map"]:
                            checkpoint = hf.get_load_ckpt(
                                hf.base_dir,
                                hf.wmap["weight_map"][srckey],
                            )
                        else:
                            checkpoint = None
                    else:
                        checkpoint = hf.get_load_ckpt(*os.path.split(hf.model_path))
                else:
                    checkpoint = None
                if checkpoint is None:
                    continue
                w = get_weight(checkpoint, srckey)
                if w is not None:
                    if srcmap is not None:
                        w = eval(
                            "w" + srcmap,
                            {
                                "w": w,
                                "hidden_size": model_config["hidden_size"],
                                "transformer_ff": model_config["transformer_ff"],
                            },
                        ).contiguous()
                    if target == "encoder.class_embedding.weight":
                        eole_safetensor[target] = w.unsqueeze(0)
                    else:
                        eole_safetensor[target] = w
            return eole_safetensor

        if shard == 0:
            eole_safetensor = build_first_shard(hf, eole_safetensor)

        # TODO: could we reverse the mapping and loop on params instead? (would reduce conditions)
        for ckpt in shard_checkpoints[shard]:
            print("Loading %s" % ckpt)
            checkpoint = hf.checkpoint(ckpt)
            for i in shard_layer_ranges[shard]:
                prefix_mapping = (
                    ("encoder", hf.encoder_layer_prefix, "encoder.transformer_layers."),
                    ("encoder.sam", hf.encoder_sam_layer_prefix, "encoder.sam.blocks."),
                    ("decoder", hf.decoder_layer_prefix, "decoder.transformer_layers."),
                )
                for section, hf_prefix, eole_prefix in prefix_mapping:
                    if hf_prefix is None:
                        continue
                    key_map = KEY_MAPS[hf.arch].get(section, {})
                    for target, source in key_map.items():
                        for param in params:
                            # TODO: this should be cleaned up when rationalizing encoder/decoder mappings
                            if not (isinstance(source, str) or isinstance(source, tuple)):
                                continue
                            srckey, srcmap = source if isinstance(source, tuple) else (source, None)

                            if srckey.endswith("."):
                                srckey = srckey + param
                            w = get_weight(
                                checkpoint,
                                hf_prefix + str(i) + srckey,
                            )

                            if w is not None:
                                if srcmap is not None:
                                    hidden_size = (
                                        model_config["hidden_size"]
                                        if section.startswith("decoder")
                                        else model_config["encoder"]["hidden_size"]
                                    )
                                    w = eval(
                                        "w" + srcmap,
                                        {
                                            "w": w,
                                            "hidden_size": hidden_size,
                                            "transformer_ff": model_config["transformer_ff"],
                                        },
                                    ).contiguous()
                                target1 = target
                                if target.endswith("."):
                                    target1 = target + param
                                eole_safetensor[eole_prefix + str(i) + target1] = w

        # Convert to another dtype if specified
        if args.dtype is not None:
            for key in eole_safetensor.keys():
                eole_safetensor[key] = eole_safetensor[key].to(TORCH_DTYPES[args.dtype])
        print("Saving output model shard: %d" % shard)
        save_file(
            eole_safetensor,
            os.path.join(args.output, "model.{:02d}.safetensors".format(shard)),
        )


def check_sentencepiece_tokenizer(hf):
    tokenizer_basename = os.path.basename(hf.tokenizer_model)
    if hf.tokenizer_json is not None:
        tokenizer_vocab = hf.tokenizer["model"]["vocab"]
        if isinstance(tokenizer_vocab, dict):
            vocab = list(hf.tokenizer["model"]["vocab"].keys())
        elif isinstance(tokenizer_vocab, list):
            vocab = [token for token, freq in tokenizer_vocab]
        else:
            raise NotImplementedError(f"Type {type(tokenizer_vocab)} is not supported for SentencePiece vocab.")
    else:
        vocab = get_sentencepiece_vocab(hf.tokenizer_model)
        if hf.tokenizer_json is not None:
            # We need to add 'added_tokens' that are not in the SP model
            newtokens = [tok["content"] for tok in hf.tokenizer["added_tokens"] if tok["content"] not in vocab]
            vocab.extend(newtokens)
            for tok in hf.tokenizer["added_tokens"]:
                vocab[tok["id"]] = tok["content"]
    src_vocab = pyonmttok.build_vocab_from_tokens(
        vocab,
    )
    return src_vocab, tokenizer_basename


def check_bpe_tokenizer(hf, vocabs, directory_path):
    config = hf.config.get("text_config", hf.config)
    vocab_size = hf.vocab_size
    # gpt2_pretok
    pretokenizers = hf.tokenizer.get("pre_tokenizer", {}).get("pretokenizers", [{}])
    pre_tokenizer = hf.tokenizer.get("pre_tokenizer", None)
    pretokenizers = pre_tokenizer.get("pretokenizers", None)
    if pretokenizers is None:
        pretokenizers = [pre_tokenizer]
    for pretokenizer in pretokenizers:
        if pretokenizer.get("type", None) == "ByteLevel":
            gpt2_pretok = True
    if "pad_token" in vocabs["specials"]:
        vocab = [tok for tok in hf.tokenizer["model"]["vocab"]]
    else:
        vocab = [
            tok if tok != "Ā" else DefaultTokens.PAD
            # "Ā" is '\x00' in unicode (cf tokenize.py gpt2 mapping)
            for tok in hf.tokenizer["model"]["vocab"]
        ]
        if DefaultTokens.PAD in vocab:
            vocabs["specials"]["pad_token"] = DefaultTokens.PAD
    voc_size = len(vocab)
    if vocab_size > voc_size:
        for i in range(vocab_size - voc_size):
            vocab.append(DefaultTokens.VOCAB_PAD + str(i))
    for tok in hf.tokenizer["added_tokens"]:
        vocab[tok["id"]] = tok["content"]
    src_vocab = pyonmttok.build_vocab_from_tokens(vocab)
    # TODO: not sure for which model(s) this is needed
    for token_name in ["bos_token", "unk_token", "eos_token", "pad_token"]:
        if f"{token_name}_id" in config.keys():
            token = config[f"{token_name}_id"]
            if isinstance(token, list):
                vocabs["specials"][token_name] = vocab[token[0]]
            elif isinstance(token, str):
                vocabs["specials"][token_name] = vocab[token]
            # elif isinstance(token, dict):
            #     vocabs["specials"][token_name] = token["content"]
    tokenizer_basename = "bpe.model"
    with open(os.path.join(directory_path, tokenizer_basename), "w", encoding="utf-8") as bpemodel:
        bpemodel.write("v3;false;false;false;Ġ;Ġ\n")
        for merge in hf.tokenizer["model"]["merges"]:
            if isinstance(merge, str):
                bpemodel.write(merge + "\n")
            elif isinstance(merge, list):
                bpemodel.write(" ".join(merge) + "\n")
            else:
                raise NotImplementedError(f"Type {type(merge)} is not supported for BPE merges.")
    return src_vocab, tokenizer_basename, vocabs, gpt2_pretok


def save_vocab(vocabs, src_vocab, directory_path):
    vocabs["src"] = src_vocab
    vocabs["tgt"] = src_vocab
    vocab_dict = vocabs_to_dict(vocabs)
    with open(os.path.join(directory_path, "vocab.json"), "w", encoding="utf-8") as f:
        json.dump(vocab_dict, f, indent=2, ensure_ascii=False)

    with open(os.path.join(directory_path, "vocab.txt"), "w", encoding="utf-8") as vocabfile:
        for tok in vocab_dict["src"]:
            vocabfile.write(tok + "\n")


@register_bin(name="HF")
class LlamaHFConverter(BaseBin):
    @classmethod
    def add_args(cls, parser):
        parser.add_argument(
            "--model_dir",
            type=str,
            required=True,
            help="""Path to the input (to convert) model directory""",
        )

        parser.add_argument(
            "--output",
            type=str,
            required=True,
            help="""Path to the output (converted) model directory""",
        )
        parser.add_argument(
            "--nshards",
            type=int,
            default=1,
            help="""Number of safetensors shards weights will be spread across.""",
        )
        parser.add_argument(
            "--token",
            type=str,
            default="",
            help="""HF token""",
        )
        parser.add_argument(
            "--dtype",
            type=str,
            default=None,
            choices=TORCH_DTYPES.keys(),
            help="Specify which dtype to save model parameters into, " "default will keep the same as the input.",
        )
        parser.add_argument(
            "--tokenizer",
            type=str,
            default="hf",
            choices=["hf", "onmt"],
            help="Specify which tokenizer should be used by default.",
        )

    @classmethod
    def run(cls, args):
        hf = HuggingfaceFiles.fetch(args)

        # Build eole compatible configs from HF config
        model_config, training_config, params = build_config_dict(hf)
        gpt2_pretok = False

        # Deduce dtype from args or config, or default to fp16
        compute_dtype = args.dtype or hf.config.get("torch_dtype") or "fp16"

        # Check tokenizer and vocab related configuration
        (
            add_bos_token,
            chat_template,
            optional_eos,
            mapped_tokens,
        ) = check_tokenizer_config(hf)
        vocabs = check_special_tokens(hf)
        # TODO: could we have a better condition to check for sentencepiece/bpe?
        if hf.tokenizer_model is not None:  # sentencepiece mode
            src_subword_type = "sentencepiece"
            src_vocab, tokenizer_basename = check_sentencepiece_tokenizer(hf)
        else:  # BPE mode - we leverage the HF tokenizer.json info
            src_subword_type = "bpe"
            src_vocab, tokenizer_basename, vocabs, gpt2_pretok = check_bpe_tokenizer(hf, vocabs, args.output)

        # Configure transforms
        match TOK_TABLE[hf.arch], args.tokenizer:
            case "huggingface_tokenize", "hf":
                transforms = ["huggingface_tokenize"]
                transforms_configs = {
                    TOK_TABLE[hf.arch]: {"max_length": 512},
                }
            case _:
                transforms = ["onmt_tokenize", "filtertoolong"]
                transforms_configs = {
                    "filtertoolong": {"src_seq_length": 512, "tgt_seq_length": 512},
                    "onmt_tokenize": {
                        "src_subword_type": src_subword_type,
                        "src_subword_model": os.path.join("${MODEL_PATH}", tokenizer_basename),
                        "gpt2_pretok": gpt2_pretok,
                        "mapped_tokens": mapped_tokens,
                    },
                }

        if hf.config.get("decoder_start_token_id", None) is not None:
            vocabs["decoder_start_token"] = src_vocab.ids_to_tokens[hf.config["decoder_start_token_id"]]
        elif add_bos_token:
            vocabs["decoder_start_token"] = vocabs["specials"]["bos_token"]
        else:
            vocabs["decoder_start_token"] = ""

        # Check HF generation config for default settings
        generation_config_dict = check_generation_config(hf)

        # Save vocab files to output model directory
        save_vocab(vocabs, src_vocab, args.output)

        # Build shards
        build_shards(model_config, hf, args, params)

        # Build eole config and save to output model directory
        config = TrainConfig(
            data=None,
            skip_empty_level="silent",
            save_data=None,
            n_sample=0,
            src_vocab=None,
            tgt_vocab=None,
            share_vocab=True,
            src_vocab_size=hf.vocab_size,
            tgt_vocab_size=hf.vocab_size,
            vocab_size_multiple=8,
            decoder_start_token=vocabs["decoder_start_token"],
            **vocabs["specials"],
            transforms=transforms,
            transforms_configs=transforms_configs,
            model=ARCH_TABLE[hf.arch](
                **model_config,
                huggingface_model=get_huggingface_model(args),
            ),
            training=TrainingConfig(
                compute_dtype=compute_dtype,
                batch_size=896,
                batch_size_multiple=1,
                batch_type="tokens",
                normalization="tokens",
                accum_count=[32],
                accum_steps=[0],
                valid_batch_size=256,
                **training_config,
            ),
        )
        # Grab only explicitly set fields to save in the model's json config
        config_dict = recursive_model_fields_set(config)

        # Add inference related settings for transparent default behaviour
        inference_dict = {
            "optional_eos": optional_eos,
            # TODO: map other settings from HF decoding_config.json
            **generation_config_dict,
            **chat_template,
        }
        config_dict["inference"] = inference_dict

        with open(os.path.join(args.output, "config.json"), "w", encoding="utf-8") as f:
            json.dump(config_dict, f, indent=2, ensure_ascii=False)



from argparse import ArgumentParser
from eole.config.run import TrainConfig, PredictConfig
from glob import glob
import os
import yaml

parser = ArgumentParser()

parser.add_argument("recipes_dir")

args = parser.parse_args()

configs_to_check = glob(os.path.join(args.recipes_dir, "*", "*.yaml"))

for config_path in configs_to_check:
    print(f"Checking: {config_path}")
    with open(config_path) as f:
        config_dict = yaml.safe_load(f)
    if "training" in config_dict.keys() or "model" in config_dict.keys():
        # must be a TrainConfig
        try:
            # pop data to prevent missing data error
            config_dict["data"] = {}
            train_config = TrainConfig(**config_dict)
        except Exception as e:
            raise e
    elif "models" in config_dict.keys():
        # server example config, ignore for now
        pass
    else:
        # must be a PredictConfig
        try:
            predict_config = PredictConfig(**config_dict)
        except Exception as e:
            raise e
